import os
import numpy as np
import torch
import torch.nn as nn
from huggingface_hub import hf_hub_download

from utils import gen_data_iter as gen_dataIter
from utils import get_config_attr
from config import (
    MASK_TOKEN, PAD_TOKEN,
    NUM_LABELS_N_MASKS, AVAILABLE_FINE_TUNING_STRATEGY,
)
from optimization_utils import (
    apply_fine_tuning_strategy_improved,
    step_optimizers
)

# Model constructors per architecture
from base_model.model_construction.phobert import get_phobert_normalizer
from base_model.model_construction.visobert import get_visobert_normalizer
from base_model.model_construction.bartpho import get_bartpho_normalizer
from base_model.model_construction.vit5 import get_vit5_normalizer


def _build_student_model(base_model, vocab_size, append_n_mask, nsw_detector, args):
    model_name = get_config_attr(args, 'model.pretrained_model_name', get_config_attr(args, 'pretrained_model_name'))
    if base_model == 'phobert':
        return get_phobert_normalizer(vocab_size, mask_n_predictor=append_n_mask, nsw_detector=nsw_detector, model_name=model_name)
    elif base_model == 'bartpho':
        return get_bartpho_normalizer(vocab_size, mask_n_predictor=append_n_mask, nsw_detector=nsw_detector, model_name=model_name)
    elif base_model == 'vit5':
        return get_vit5_normalizer(vocab_size, mask_n_predictor=append_n_mask, nsw_detector=nsw_detector, model_name=model_name)
    elif base_model == 'visobert':
        return get_visobert_normalizer(vocab_size, mask_n_predictor=append_n_mask, nsw_detector=nsw_detector, model_name=model_name)
    else:
        raise ValueError(f"Unsupported base_model: {base_model}")


def get_label_n_masks(input, num_labels_n_masks):
    output = torch.empty_like(input).long()
    for ind_sent in range(input.size(0)):
        count = 0
        for ind_word in range(input.size(1)):
            if input[ind_sent, ind_word] == 1:
                output[ind_sent, ind_word] = -1
                if count == 0:
                    ind_multi_bpe = ind_word - 1
                count += 1
            elif input[ind_sent, ind_word] == 0:
                if ind_word > 0 and input[ind_sent, ind_word -1] == 1:
                    output[ind_sent, ind_multi_bpe] = min(count, num_labels_n_masks-1)
                    count = 0
                output[ind_sent, ind_word] = 0
    return output


class ViSoNormalizerTrainer:
    def __init__(self, args, tokenizer, logger=None):
        self.args = args
        self.name = get_config_attr(args, 'model.base_model', get_config_attr(args, 'base_model'))
        self.logger = logger
        self.tokenizer = tokenizer
        self.manual_seed = get_config_attr(args, 'system.seed', get_config_attr(args, 'seed', 42))
        training_mode = get_config_attr(args, 'model.training_mode', get_config_attr(args, 'training_mode', 'weakly_supervised'))
        default_logdir = './experiments/{}/{}'.format(self.name, training_mode)
        self.model_dir = get_config_attr(args, 'paths.logdir', get_config_attr(args, 'logdir', default_logdir))
        self.sup_batch_size = get_config_attr(args, 'training.train_batch_size', get_config_attr(args, 'train_batch_size', 16))
        self.unsup_batch_size = get_config_attr(args, 'training.unsup_batch_size', get_config_attr(args, 'unsup_batch_size', 128))
        self.eval_batch_size = get_config_attr(args, 'training.eval_batch_size', get_config_attr(args, 'eval_batch_size', 128))
        self.sup_epochs = get_config_attr(args, 'training.num_epochs', get_config_attr(args, 'num_epochs', 10))
        self.unsup_epochs = get_config_attr(args, 'training.num_unsup_epochs', get_config_attr(args, 'num_unsup_epochs', 5))
        self.device = get_config_attr(args, 'system.device', get_config_attr(args, 'device', 'cuda'))
        self.use_gpu = self.device=='cuda'
        self.append_n_mask = get_config_attr(args, 'model.append_n_mask', get_config_attr(args, 'append_n_mask', False))
        self.nsw_detector = get_config_attr(args, 'model.nsw_detector', get_config_attr(args, 'nsw_detector', False))
        self.pretrained_model_name = get_config_attr(args, 'model.pretrained_model_name', get_config_attr(args, 'pretrained_model_name', None))

        self.model = _build_student_model(self.name, len(self.tokenizer), self.append_n_mask, self.nsw_detector, args)
        if self.use_gpu:
            self.model = self.model.to(self.device)
        self.fine_tuning_strategy = get_config_attr(args, 'model.fine_tuning_strategy', get_config_attr(args, 'fine_tuning_strategy', 'flexible_lr'))
        self.learning_rate = get_config_attr(args, 'model.learning_rate', get_config_attr(args, 'learning_rate', 1e-3))
        self.loss_weights = get_config_attr(args, 'model.loss_weights', get_config_attr(args, 'loss_weights', False))
        self.soft_labels = get_config_attr(args, 'model.soft_labels', get_config_attr(args, 'soft_labels', False))

    def _truncate_and_build_masks(self, input_tokens_tensor, output_tokens_tensor=None):
        if hasattr(self.model, 'roberta'):
            cfg_max = int(getattr(self.model.roberta.config, 'max_position_embeddings', input_tokens_tensor.size(1)))
            tbl_max = int(getattr(self.model.roberta.embeddings.position_embeddings, 'num_embeddings', cfg_max))
            max_pos = min(cfg_max, tbl_max)
            eff_max = max(1, max_pos - 2)
            if input_tokens_tensor.size(1) > eff_max:
                input_tokens_tensor = input_tokens_tensor[:, :eff_max]
                if output_tokens_tensor is not None and output_tokens_tensor.dim() == 2 and output_tokens_tensor.size(1) > eff_max:
                    output_tokens_tensor = output_tokens_tensor[:, :eff_max]
            pad_id_model = getattr(self.model.roberta.config, 'pad_token_id', None)
            if pad_id_model is None:
                pad_id_model = getattr(self.model.roberta.embeddings.word_embeddings, 'padding_idx', None)
            if pad_id_model is None:
                pad_id_model = self.tokenizer.pad_token_id if self.tokenizer.pad_token_id is not None else 1
            input_mask = (input_tokens_tensor != pad_id_model).long()
            token_type_ids = torch.zeros_like(input_tokens_tensor)
            return input_tokens_tensor, output_tokens_tensor, token_type_ids, input_mask
        # bart branch
        pad_id_model = 1
        input_mask = torch.ones_like(input_tokens_tensor)
        token_type_ids = None
        return input_tokens_tensor, output_tokens_tensor, token_type_ids, input_mask

    def epoch_run(self, batchIter, mode, epoch, n_epochs, optimizer=None, scheduler=None):
        mask_id = self.tokenizer.convert_tokens_to_ids([MASK_TOKEN])[0]
        pad_id = self.tokenizer.convert_tokens_to_ids([PAD_TOKEN])[0]

        labels_n_mask_prediction = None
        loss = 0
        loss_norm = 0
        loss_n_masks_pred = 0
        loss_nsw_detection = 0
        num_batch = 0 if mode != 'train_pseudo' else len(batchIter['id'])
        if mode != 'train_pseudo':
            while True:
                try:
                    batch = batchIter.__next__()
                    num_batch += 1
                    input_ids = batch['input_ids']
                    input_tokens_tensor = torch.LongTensor(input_ids)
                    output_ids = batch['output_ids']
                    output_tokens_tensor = torch.LongTensor(output_ids)

                    input_tokens_tensor, output_tokens_tensor, token_type_ids, input_mask = self._truncate_and_build_masks(input_tokens_tensor, output_tokens_tensor)
                    if self.use_gpu:
                        input_tokens_tensor = input_tokens_tensor.cuda()
                        output_tokens_tensor = output_tokens_tensor.cuda()
                        input_mask = input_mask.cuda()
                        if token_type_ids is not None:
                            token_type_ids = token_type_ids.cuda()

                    if hasattr(self.model, 'roberta'):
                        vocab_size = self.model.roberta.config.vocab_size
                        if torch.any(input_tokens_tensor < 0) or torch.any(input_tokens_tensor >= vocab_size):
                            unk_id = self.tokenizer.unk_token_id if self.tokenizer.unk_token_id is not None else 0
                            input_tokens_tensor = input_tokens_tensor.clamp(min=0, max=vocab_size - 1)
                            input_tokens_tensor[input_tokens_tensor >= vocab_size] = unk_id

                    if self.append_n_mask:
                        labels_n_mask_prediction = get_label_n_masks(input_tokens_tensor == mask_id, NUM_LABELS_N_MASKS)
                        assert (((input_tokens_tensor == mask_id).nonzero() == (labels_n_mask_prediction == -1).nonzero())).all()
                        labels_n_mask_prediction[input_tokens_tensor == pad_id] = -1

                    if self.nsw_detector:
                        standard_labels = (input_tokens_tensor != output_tokens_tensor).long()
                        standard_labels[input_tokens_tensor == pad_id] = -1
                    else:
                        standard_labels = None

                    feeding_the_model_with_label = output_tokens_tensor.clone()

                    if optimizer is not None:
                        portion_mask = min((1- (epoch + 1) / n_epochs), 0.6)
                        mask_normed = np.random.random() < portion_mask
                        if mask_normed:
                            feeding_the_model_with_label[input_tokens_tensor == output_tokens_tensor] = -1
                            if np.random.random() < 0.5:
                                input_tokens_tensor[input_tokens_tensor != output_tokens_tensor] = mask_id

                    if hasattr(self.model, 'roberta'):
                        loss_dic, logits, _ = self.model(input_tokens_tensor, token_type_ids, input_mask,
                                                         labels=feeding_the_model_with_label,
                                                         labels_n_masks=labels_n_mask_prediction,
                                                         standard_labels=standard_labels)
                    else:
                        loss_dic, logits, _ = self.model(input_tokens_tensor, input_mask,
                                                         labels=feeding_the_model_with_label,
                                                         labels_n_masks=labels_n_mask_prediction,
                                                         standard_labels=standard_labels)
                    _loss = loss_dic["loss"]
                    loss_norm += loss_dic["loss_norm"].detach()
                    if self.append_n_mask and not isinstance(loss_dic.get("loss_n_masks_pred", 0), int):
                        loss_n_masks_pred += loss_dic["loss_n_masks_pred"].detach()
                    if self.nsw_detector and not isinstance(loss_dic.get("loss_nsw_detection", 0), int):
                        loss_nsw_detection += loss_dic["loss_nsw_detection"].detach()

                    loss += _loss.detach()
                    if optimizer is not None:
                        _loss.backward()
                        step_optimizers(
                            optimizer,
                            scheduler=scheduler,
                            model=self.model,
                            max_grad_norm=getattr(self.args, 'model.max_grad_norm', 1.0)
                        )
                except StopIteration:
                    print("BREAKING ITERATION")
                    break
        else:
            for i in range(num_batch):
                input_ids = batchIter['input_ids'][i]
                input_tokens_tensor = torch.LongTensor(input_ids)
                output_tokens_tensor = torch.tensor(batchIter['proba'][i]) if self.soft_labels else torch.LongTensor(batchIter['labels'][i])

                input_tokens_tensor, output_tokens_tensor, token_type_ids, input_mask = self._truncate_and_build_masks(input_tokens_tensor, output_tokens_tensor)
                if self.use_gpu:
                    input_tokens_tensor = input_tokens_tensor.cuda()
                    output_tokens_tensor = output_tokens_tensor.cuda()
                    input_mask = input_mask.cuda()
                    if token_type_ids is not None:
                        token_type_ids = token_type_ids.cuda()

                if hasattr(self.model, 'roberta'):
                    vocab_size = self.model.roberta.config.vocab_size
                    if torch.any(input_tokens_tensor < 0) or torch.any(input_tokens_tensor >= vocab_size):
                        unk_id = self.tokenizer.unk_token_id if self.tokenizer.unk_token_id is not None else 0
                        input_tokens_tensor = input_tokens_tensor.clamp(min=0, max=vocab_size - 1)
                        input_tokens_tensor[input_tokens_tensor >= vocab_size] = unk_id

                if self.append_n_mask:
                    labels_n_mask_prediction = get_label_n_masks(input_tokens_tensor == mask_id, NUM_LABELS_N_MASKS)
                    assert (((input_tokens_tensor == mask_id).nonzero() == (labels_n_mask_prediction == -1).nonzero())).all()
                    labels_n_mask_prediction[input_tokens_tensor == pad_id] = -1

                if self.nsw_detector:
                    if input_tokens_tensor.shape != output_tokens_tensor.shape:
                        standard_labels = torch.zeros_like(input_tokens_tensor, dtype=torch.long)
                        standard_labels.fill_(1)
                    else:
                        standard_labels = (input_tokens_tensor != output_tokens_tensor).long()
                    standard_labels[input_tokens_tensor == pad_id] = -1
                else:
                    standard_labels = None

                feeding_the_model_with_label = output_tokens_tensor.clone()

                sample_weights = None
                if self.loss_weights:
                    sample_weights = batchIter['weight'][i]
                    sample_weights = torch.tensor(sample_weights)
                    if self.use_gpu:
                        sample_weights = sample_weights.cuda()

                if hasattr(self.model, 'roberta'):
                    loss_dic, logits, _ = self.model(input_tokens_tensor, token_type_ids, input_mask,
                                                    labels=feeding_the_model_with_label,
                                                    labels_n_masks=labels_n_mask_prediction,
                                                    standard_labels=standard_labels,
                                                    sample_weights=sample_weights,
                                                    soft_labels=self.soft_labels)
                else:
                    loss_dic, logits, _ = self.model(input_tokens_tensor, input_mask,
                                                    labels=feeding_the_model_with_label,
                                                    labels_n_masks=labels_n_mask_prediction,
                                                    standard_labels=standard_labels,
                                                    sample_weights=sample_weights,
                                                    soft_labels=self.soft_labels)
                _loss = loss_dic["loss"]
                loss_norm += loss_dic["loss_norm"].detach()
                if self.append_n_mask and not isinstance(loss_dic.get("loss_n_masks_pred", 0), int):
                    loss_n_masks_pred += loss_dic["loss_n_masks_pred"].detach()
                if self.nsw_detector and not isinstance(loss_dic.get("loss_nsw_detection", 0), int):
                    loss_nsw_detection += loss_dic["loss_nsw_detection"].detach()

                loss += _loss.detach()
                _loss.backward()
                step_optimizers(
                    optimizer,
                    scheduler=scheduler,
                    model=self.model,
                    max_grad_norm=getattr(self.args, 'model.max_grad_norm', 1.0)
                )

        return loss/num_batch

    def train(self, train_data, dev_data=None):
        train_sent_len_ls = list(set(train_data['sent_len']))
        dev_sent_len_ls = list(set(dev_data['sent_len']))
        best_val_loss = np.inf
        train_losses = []
        dev_losses = []

        optimizer, scheduler = apply_fine_tuning_strategy_improved(
            self.fine_tuning_strategy,
            self.model,
            self.learning_rate,
            self.args,
            self.append_n_mask,
            self.nsw_detector
        )

        for epoch in range(self.sup_epochs):
            trainIter = gen_dataIter(train_data, self.sup_batch_size, train_sent_len_ls, shuffle=True, seed=self.manual_seed)
            devIter = gen_dataIter(dev_data, self.sup_batch_size, dev_sent_len_ls)

            self.model.train()
            loss_train = self.epoch_run(trainIter, "train", epoch=epoch+1, n_epochs=self.sup_epochs, optimizer=optimizer, scheduler=scheduler)
            train_losses.append(loss_train.item())
            self.model.eval()
            with torch.no_grad():
                loss_dev = self.epoch_run(devIter, "dev", epoch=epoch+1, n_epochs=self.sup_epochs)
                dev_losses.append(loss_dev.item())
                if loss_dev < best_val_loss:
                    best_model = self.model
                    best_val_loss = loss_dev
                    train_loss = loss_train
            if self.logger:
                self.logger.info("EPOCH {}/{}: train_loss: {} - val_loss: {} - best_val_loss: {}".format(epoch+1, self.sup_epochs, loss_train, loss_dev, best_val_loss))
        self.model = best_model
        return {
            'train_loss': train_losses,
            'dev_loss': dev_losses,
            'best_dev_loss': best_val_loss.item()
        }

    def finetune(self, train_data, dev_data=None):
        train_sent_len_ls = list(set(train_data['sent_len']))
        dev_sent_len_ls = list(set(dev_data['sent_len']))
        best_val_loss = np.inf
        train_losses = []
        dev_losses = []

        optimizer, scheduler = apply_fine_tuning_strategy_improved(
            self.fine_tuning_strategy,
            self.model,
            self.learning_rate,
            self.args,
            self.append_n_mask,
            self.nsw_detector
        )

        for epoch in range(self.sup_epochs):
            trainIter = gen_dataIter(train_data, self.sup_batch_size, train_sent_len_ls, shuffle=True)
            devIter = gen_dataIter(dev_data, self.sup_batch_size, dev_sent_len_ls)

            self.model.train()
            loss_train = self.epoch_run(trainIter, "finetune", epoch=epoch+1, n_epochs=self.sup_epochs, optimizer=optimizer, scheduler=scheduler)
            train_losses.append(loss_train.item())
            self.model.eval()
            with torch.no_grad():
                loss_dev = self.epoch_run(devIter, "dev", epoch=epoch+1, n_epochs=self.sup_epochs)
                dev_losses.append(loss_dev.item())
                if loss_dev < best_val_loss:
                    best_model = self.model
                    best_val_loss = loss_dev
                    train_loss = loss_train
            if self.logger:
                self.logger.info("EPOCH {}/{}: train_loss: {} - val_loss: {} - best_val_loss: {}".format(epoch+1, self.sup_epochs, loss_train, loss_dev, best_val_loss))
        self.model = best_model
        return {
            'train_loss': train_losses,
            'dev_loss': dev_losses,
            'best_dev_loss': best_val_loss.item()
        }

    def train_pseudo(self, train_data, dev_data=None):
        dev_sent_len_ls = list(set(dev_data['sent_len']))
        best_val_loss = np.inf
        train_losses = []
        dev_losses = []

        optimizer, scheduler = apply_fine_tuning_strategy_improved(
            self.fine_tuning_strategy,
            self.model,
            self.learning_rate,
            self.args,
            self.append_n_mask,
            self.nsw_detector
        )

        for epoch in range(self.unsup_epochs):
            devIter = gen_dataIter(dev_data, self.unsup_batch_size, dev_sent_len_ls)
            self.model.train()
            loss_train = self.epoch_run(train_data, "train_pseudo", epoch=epoch+1, n_epochs=self.unsup_epochs, optimizer=optimizer, scheduler=scheduler)
            train_losses.append(loss_train.item())
            self.model.eval()
            with torch.no_grad():
                loss_dev = self.epoch_run(devIter, "dev", epoch=epoch+1, n_epochs=self.unsup_epochs)
                dev_losses.append(loss_dev.item())
                if loss_dev < best_val_loss:
                    best_model = self.model
                    best_val_loss = loss_dev
                    train_loss = loss_train
            if self.logger:
                self.logger.info("EPOCH {}/{}: train_loss: {} - val_loss: {} - best_val_loss: {}".format(epoch+1, self.unsup_epochs, loss_train, loss_dev, best_val_loss))
        self.model = best_model
        return {
            'train_loss': train_losses,
            'dev_loss': dev_losses,
            'best_dev_loss': best_val_loss.item()
        }

    def predict(self, data, inference_mode, dataIter=None):
        label = False
        if data is not None:
            len_ls = list(set(data['sent_len']))
            dataIter = gen_dataIter(data, batch_size=self.eval_batch_size, len_list=len_ls)
        preds = []
        probas = []
        features = []
        rule_pred = []
        sent_ids = []
        input_ids = []
        output_ids = []
        align_index = []

        while True:
            try:
                with torch.no_grad():
                    batch = dataIter.__next__()
                    input_tokens_tensor = batch['input_ids']
                    input_tokens_tensor = torch.LongTensor(input_tokens_tensor)
                    token_type_ids = torch.zeros_like(input_tokens_tensor) if hasattr(self.model, 'roberta') else None
                    input_mask = torch.ones_like(input_tokens_tensor)
                    if hasattr(self.model, 'roberta'):
                        input_tokens_tensor, _, token_type_ids, input_mask = self._truncate_and_build_masks(input_tokens_tensor)
                    if self.use_gpu:
                        input_tokens_tensor = input_tokens_tensor.cuda()
                        input_mask = input_mask.cuda()
                        if token_type_ids is not None:
                            token_type_ids = token_type_ids.cuda()
                    self.model.eval()
                    if hasattr(self.model, 'roberta'):
                        _, logits, feature = self.model(input_tokens_tensor, token_type_ids, input_mask)
                    else:
                        _, logits, feature = self.model(input_tokens_tensor, input_mask)

                    pred = torch.argmax(logits["logits_norm"], dim=-1)
                    if not inference_mode:
                        proba = torch.softmax(logits["logits_norm"], dim=-1)

                    preds.append(pred.detach().cpu().numpy())
                    if not inference_mode:
                        probas.append(proba.detach().cpu().numpy())
                        features.append(feature.detach().cpu().numpy())
                        rule_pred.append(np.array(batch['weak_labels']))
                        if 'output_ids' in batch:
                            label = True
                            output_ids.append(np.array(batch['output_ids']))
                    sent_ids.append(batch['id'])
                    input_ids.append(np.array(batch['input_ids']))
                    align_index.append(batch['align_index'])
            except StopIteration:
                print("BREAKING DATA ITERATION")
                break

        if inference_mode:
            return {
                "id": sent_ids,
                "input_ids": input_ids,
                "align_index": align_index,
                "preds":  preds,
            }

        return {
            "id": sent_ids,
            "input_ids": input_ids,
            "output_ids": output_ids if label else None,
            "align_index": align_index,
            "preds":  preds,
            "proba": probas,
            "features": features,
            "weak_labels": rule_pred,
        }

    def inference(self, user_input):
        inputs = self.tokenizer(user_input, return_tensors="pt")
        self.model.eval()
        with torch.no_grad():
            if self.use_gpu:
                input_tokens_tensor = inputs.input_ids.cuda()
                input_mask = inputs.attention_mask.cuda()
            else:
                input_tokens_tensor = inputs.input_ids
                input_mask = inputs.attention_mask
            if hasattr(self.model, 'roberta'):
                token_type_ids = torch.zeros_like(input_tokens_tensor)
                if self.use_gpu:
                    token_type_ids = token_type_ids.cuda()
                _, logits, feature = self.model(input_tokens_tensor, token_type_ids, input_mask)
            else:
                _, logits, feature = self.model(input_tokens_tensor, input_mask)
            pred = torch.argmax(logits["logits_norm"], dim=-1).squeeze()
        return pred.cpu().tolist()

    def load(self, savefolder):
        # Prefer HuggingFace convention
        model_file = os.path.join(savefolder, "pytorch_model.bin")
        if not os.path.exists(model_file):
            # Fallback to legacy filename
            model_file = os.path.join(savefolder, "final_model.pt")
        if self.logger:
            self.logger.info("Loading student from {}".format(model_file))
        self.model.load_state_dict(torch.load(model_file))
        return

    def save(self, savefolder):
        # Save using HuggingFace filename convention
        model_file = os.path.join(savefolder, "pytorch_model.bin")
        if self.logger:
            self.logger.info("Saving model at {}".format(model_file))
        torch.save(self.model.state_dict(), model_file)
        self._save_tokenizer(savefolder)
        # Save minimal HF config.json by downloading base config and applying overrides
        self._save_minimal_config(savefolder)
        
    def _save_tokenizer(self, savefolder):
        if hasattr(self, 'tokenizer') and self.tokenizer is not None:
            self.tokenizer.save_pretrained(savefolder)
        if self.logger:
            self.logger.info("Saved tokenizer at {}".format(savefolder))

    def _save_minimal_config(self, savefolder):
        try:
            os.makedirs(savefolder, exist_ok=True)
            base_cfg = {}
            if self.pretrained_model_name:
                try:
                    cfg_path = hf_hub_download(repo_id=self.pretrained_model_name, filename="config.json")
                    import json
                    with open(cfg_path, "r", encoding="utf-8") as f:
                        base_cfg = json.load(f)
                except Exception as e:
                    if self.logger:
                        self.logger.warning("Failed to download base config.json from {}: {}".format(self.pretrained_model_name, e))

            # Apply minimal overrides
            updated = dict(base_cfg)
            updated['vocab_size'] = int(len(self.tokenizer))
            updated['mask_n_predictor'] = bool(self.append_n_mask)
            updated['nsw_detector'] = bool(self.nsw_detector)

            import json
            with open(os.path.join(savefolder, "config.json"), "w", encoding="utf-8") as f:
                json.dump(updated, f, indent=2)
            if self.logger:
                self.logger.info("Saved minimal HuggingFace config.json at {}".format(os.path.join(savefolder, "config.json")))
        except Exception as e:
            if self.logger:
                self.logger.warning("Could not save minimal config.json: {}".format(e))


import re
from emot.emo_unicode import UNICODE_EMOJI

def regex_or(*items):
    return '(?:' + '|'.join(items) + ')'


punctChars = r"['\"""''.?!…,:;]"
#punctSeq   = punctChars+"+"	#'anthem'. => ' anthem '.
punctSeq   = r"['\"""'']+|[.?!,…]+|[:;]+"	#'anthem'. => ' anthem ' .
entity     = r"&(?:amp|lt|gt|quot);"

#URLs
urlStart1  = r"(?:https?://|\bwww\.)"
commonTLDs = r"(?:com|org|edu|gov|net|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|pro|tel|travel|xxx)"
ccTLDs	 = r"(?:ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|" + \
r"bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|" + \
r"er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|" + \
r"hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|" + \
r"lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|" + \
r"nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|" + \
r"sl|sm|sn|so|sr|ss|st|su|sv|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|" + \
r"va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|za|zm|zw)"	#TODO: remove obscure country domains?
urlStart2  = r"\b(?:[A-Za-z\d-])+(?:\.[A-Za-z0-9]+){0,3}\." + regex_or(commonTLDs, ccTLDs) + r"(?:\."+ccTLDs+r")?(?=\W|$)"
urlBody    = r"(?:[^\.\s<>][^\s<>]*?)?"
urlExtraCrapBeforeEnd = regex_or(punctChars, entity) + "+?"
urlEnd     = r"(?:\.\.+|[<>]|\s|$)"
url        = regex_or(urlStart1, urlStart2) + urlBody + "(?=(?:"+urlExtraCrapBeforeEnd+")?"+urlEnd+")"

# Numeric
timeLike   = r"\d+(?::\d+){1,2}"  # (1:00:00)
numdotNum = r"\d+\.\d+"
rationum = r"\d+\-\d+"
numberWithCommas = r"(?:(?<!\d)\d{1,3},)+?\d{3}" + r"(?=(?:[^,\d]|$))"  #(1,5  2,5)

#currency
numComb	 = u"[\u0024\u058f\u060b\u09f2\u09f3\u09fb\u0af1\u0bf9\u0e3f\u17db\ua838\ufdfc\ufe69\uff04\uffe0\uffe1\uffe5\uffe6\u00a2-\u00a5\u20a0-\u20b9]?\\d+(?:\\.\\d+)+%?"

boundaryNotDot = regex_or("$", r"\s", r"[""\"?!,:;]", entity)
separators  = "(?:--+|―|—|~|–|=)"
decorations = u"(?:[♫♪]+|[★☆]+|[♥❤♡]+|[\u2639-\u263b]+|[\ue001-\uebbb]+)"
thingsThatSplitWords = r"[^\s\.,?\"]"

# Filename  or Word separated by dot
filename = r'\b[\w-]+\.(?:txt|pdf|docx|jpeg|jpg|png|gif|csv|xls|xlsx)\b' #(filename.txt, name.jpg)
word_separate_by_dot = r'(?<!\.\s)[a-zA-Z0-9]+(?:\.[a-zA-Z0-9]+){2,}' # (U.I.T)

#  Emoticons
# myleott: in Python the (?iu) flags affect the whole expression
#normalEyes = "(?iu)[:=]" # 8 and x are eyes but cause problems
normalEyes = "[:=]" # 8 and x are eyes but cause problems
wink = "[;]"
noseArea = "(?:|-|[^a-zA-Z0-9 ])" # doesn't get :'-(
happyMouths = r"[D\)\]\}\>]+"
sadMouths = r"[\(\[\{\<]+"
tongue = "[pPd3*]+"
otherMouths = r"(?:[oO]+|[/\\]+|[vV]+|[Ss]+|[|]+)" # remove forward slash if http://'s aren't cleaned
bfLeft = u"(♥|0|[oO]|°|[vV]|\\$|[tT]|[xX]|;|\u0ca0|@|ʘ|•|・|◕|\\^|¬|\\*)"
bfCenter = r"(?:[\.]|[_-]+)"
bfRight = r"\2"
s3 = r"(?:--['\"])"
s4 = r"(?:<|&lt;|>|&gt;)[\._-]+(?:<|&lt;|>|&gt;)"
s5 = "(?:[.][_]+[.])"

double_eyes = r"(?:[^]" + r"[^])"
basicface = "(?:" +bfLeft+bfCenter+bfRight+ ")|" +s3+ "|" +s4+ "|" + s5
eeLeft = r"[＼\\ƪԄ\(（<>;ヽ\-=~\*]+"
eeRight= u"[\\-=\\);'\u0022<>ʃ）/／ノﾉ丿╯σっµ~\\*]+"
eeSymbol = r"[^A-Za-z0-9\s\(\)\*:=-]"
eastEmote = eeLeft + "(?:"+basicface+"|" +eeSymbol+")+" + eeRight
oOEmote = r"(?:[oO]" + bfCenter + r"[oO])"

Hearts = "(?:<+/?3+)+"
Arrows = regex_or(r"(?:<*[-―—=]*>+|<+[-―—=]*>*)", u"[\u2190-\u21ff]+")
Hashtag = "#[a-zA-Z0-9_ÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚĂĐĨŨƠàáâãèéêìíòóôõùúăđĩũơƯĂẠẢẤẦẨẪẬẮẰẲẴẶẸẺẼỀỀỂưăạảấầẩẫậắằẳẵặẹẻẽềềểỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪễếệỉịọỏốồổỗộớờởỡợụủứừỬỮỰỲỴÝỶỸửữựỳỵỷỹ]+"
AtMention = "[@＠][a-zA-Z0-9_ÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚĂĐĨŨƠàáâãèéêìíòóôõùúăđĩũơƯĂẠẢẤẦẨẪẬẮẰẲẴẶẸẺẼỀỀỂưăạảấầẩẫậắằẳẵặẹẻẽềềểỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪễếệỉịọỏốồổỗộớờởỡợụủứừỬỮỰỲỴÝỶỸửữựỳỵỷỹ]+"
Bound = r"(?:\W|^|$)"
Email = regex_or("(?<=(?:\W))", "(?<=(?:^))") + r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,4}(?=" +Bound+")"

# Edge punctuation
edgePunctChars    = u"'\"\"''«»{}\\(\\)\\[\\]\\*&" #add \\p{So}? (symbols)
edgePunct    = "[" + edgePunctChars + "]"
notEdgePunct = "[a-zA-Z0-9_ÀÁÂÃÈÉÊÌÍÒÓÔÕÙÚĂĐĨŨƠàáâãèéêìíòóôõùúăđĩũơƯĂẠẢẤẦẨẪẬẮẰẲẴẶẸẺẼỀỀỂưăạảấầẩẫậắằẳẵặẹẻẽềềểỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪễếệỉịọỏốồổỗộớờởỡợụủứừỬỮỰỲỴÝỶỸửữựỳỵỷỹ]" # content characters
offEdge = r"(^|$|:|;|\s|\.|,)"  # colon here gets "(hello):" ==> "( hello ):"
EdgePunctLeft  = re.compile(offEdge + "("+edgePunct+"+)("+notEdgePunct+")", re.UNICODE)
EdgePunctRight = re.compile("("+notEdgePunct+")("+edgePunct+"+)" + offEdge, re.UNICODE)

# Datetime
# m/d/yyyy + mm/dd/yyyy
d1 = r"^(1[0-2]|0?[1-9])/(3[01]|[12][0-9]|0?[1-9])/(?:[0-9]{2})?[0-9]{2}$"
# d/m/yyyy + dd/mm/yyyy
d2 = r"^(3[01]|[12][0-9]|0?[1-9])/(1[0-2]|0?[1-9])/(?:[0-9]{2})?[0-9]{2}$"
# m/d or mm/dd
d3 = r"^(1[0-2]|0?[1-9])/(3[01]|[12][0-9]|0?[1-9])"
# d/m or dd/mm
d4 = r"^(3[01]|[12][0-9]|0?[1-9])/(1[0-2]|0?[1-9])"

date = regex_or(d1,d2,d3,d4)

emoticon = regex_or(
        # Standard version  :) :( :] :D :P
        "(?:>|&gt;)?" + regex_or(normalEyes, wink) + regex_or(noseArea,"[Oo]") + regex_or(tongue+r"(?=\W|$|RT|rt|Rt)", otherMouths+r"(?=\W|$|RT|rt|Rt)", sadMouths, happyMouths),

        # reversed version (: D:  use positive lookbehind to remove "(word):"
        # because eyes on the right side is more ambiguous with the standard usage of : ;
        regex_or("(?<=(?: ))", "(?<=(?:^))") + regex_or(sadMouths,happyMouths,otherMouths) + noseArea + regex_or(normalEyes, wink) + "(?:<|&lt;)?",

        #inspired by http://en.wikipedia.org/wiki/User:Scapler/emoticons#East_Asian_style
        eastEmote.replace("2", "1", 1), basicface,
        # iOS 'emoji' characters (some smileys, some symbols) [\ue001-\uebbb]
        # TODO should try a big precompiled lexicon from Wikipedia, Dan Ramage told me (BTO) he does this

        # myleott: o.O and O.o are two of the biggest sources of differences
        #          between this and the Java version. One little hack won't hurt...
        oOEmote
)

emoji_pattern = re.compile(
    "["
    "\U0001F600-\U0001F64F"  # emoticons
    "\U0001F300-\U0001F5FF"  # symbols & pictographs
    "\U0001F680-\U0001F6FF"  # transport & map symbols
    "\U0001F700-\U0001F77F"  # alchemical symbols
    "\U0001F780-\U0001F7FF"  # Geometric Shapes Extended
    "\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
    "\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
    "\U0001FA00-\U0001FA6F"  # Chess Symbols
    "\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
    "\U00002702-\U000027B0"  # Dingbats
    "\U000024C2-\U0001F251"
    "]+", flags=re.UNICODE)


tone_dict_map = {
    "òa": "oà",  "Òa": "Oà",  "ÒA": "OÀ",
    "óa": "oá",  "Óa": "Oá",  "ÓA": "OÁ",
    "ỏa": "oả",  "Ỏa": "Oả",  "ỎA": "OẢ",
    "õa": "oã",  "Õa": "Oã",  "ÕA": "OÃ",
    "ọa": "oạ",  "Ọa": "Oạ",  "ỌA": "OẠ",
    "òe": "oè",  "Òe": "Oè",  "ÒE": "OÈ",
    "óe": "oé",  "Óe": "Oé",  "ÓE": "OÉ",
    "ỏe": "oẻ",  "Ỏe": "Oẻ",  "ỎE": "OẺ",
    "õe": "oẽ",  "Õe": "Oẽ",  "ÕE": "OẼ",
    "ọe": "oẹ",  "Ọe": "Oẹ",  "ỌE": "OẸ",
    "ùy": "uỳ",  "Ùy": "Uỳ",  "ÙY": "UỲ",
    "úy": "uý",  "Úy": "Uý",  "ÚY": "UÝ",
    "ủy": "uỷ",  "Ủy": "Uỷ",  "ỦY": "UỶ",
    "ũy": "uỹ",  "Ũy": "Uỹ",  "ŨY": "UỸ",
    "ụy": "uỵ",  "Ụy": "Uỵ",  "ỤY": "UỴ",
    }

emoji_list = list(UNICODE_EMOJI)

Protected  = re.compile(
    regex_or(
        Hearts,
        url,
        Email,
        timeLike,
        date,
        numdotNum,
        rationum,
        numberWithCommas,
        filename,
        word_separate_by_dot,
        numComb,
        emoticon,
        Arrows,
        entity,
        punctSeq,
        separators,
        decorations,
        Hashtag,
        AtMention,
        emoji_pattern.pattern), re.UNICODE)

Emoji_Protected = re.compile(
    regex_or(
        Hearts,
        emoticon,
        Arrows,
        entity,
        decorations,
        Hashtag,
        AtMention,
        emoji_pattern.pattern), re.UNICODE)




import torch
from contextlib import nullcontext
from visonorm.utils import suppress_output, clear_console
from transformers import AutoTokenizer, AutoModelForSequenceClassification

class EmotionRecognition:
    MODEL_MAP = {
        "phobert-v1": "visolex/emotion-phobert-v1",
        "phobert-v2": "visolex/emotion-phobert-v2",        
        "xlm-roberta": "visolex/emotion-xlm-roberta",
        "bartpho": "visolex/emotion-bartpho",
        "mbert": "visolex/emotion-mbert",
        "vit5": "visolex/emotion-vit5",
        "visobert": "visolex/emotion-visobert",
        "sphobert": "visolex/emotion-sphobert",
        "roberta-gru": "visolex/emotion-roberta-gru",
        #"textcnn": "visolex/emotion-textcnn",
        #"bilstm": "visolex/emotion-bilstm",
    }
    
    @staticmethod
    def list_models():
        """
        List all available models for emotion recognition.
        Returns:
            dict: Dictionary with model names as keys and HuggingFace paths as values.
        """
        return EmotionRecognition.MODEL_MAP.copy()
    
    @staticmethod
    def list_model_names():
        """
        List all available model names for emotion recognition.
        Returns:
            list: List of available model names.
        """
        return list(EmotionRecognition.MODEL_MAP.keys())

    def __init__(self, model_name, silent=True, clear_output=True):
        if model_name not in self.MODEL_MAP:
            available = self.list_model_names()
            raise ValueError(f"Model '{model_name}' not found. Available models: {available}")
        model_path = self.MODEL_MAP[model_name]
        ctx = suppress_output() if silent else nullcontext()
        with ctx:
            try:
                self.tokenizer = AutoTokenizer.from_pretrained(model_path, trust_remote_code=True)
                self.model = AutoModelForSequenceClassification.from_pretrained(model_path, trust_remote_code=True)
            except Exception as e:
                # Try without trust_remote_code if it fails
                try:
                    self.tokenizer = AutoTokenizer.from_pretrained(model_path)
                    self.model = AutoModelForSequenceClassification.from_pretrained(model_path)
                except Exception as e2:
                    raise RuntimeError(f"Failed to load model '{model_name}' from '{model_path}'. "
                                     f"Error: {str(e2)}. Please check if the model exists on HuggingFace Hub.")
        if clear_output:
            clear_console()
        self.label_map = {
            0: "Anger", 1: "Disgust", 2: "Enjoyment", 3: "Fear",
            4: "Other", 5: "Sadness", 6: "Surprise"
        }

    def predict(self, text):
        """
        Predict emotion labels for the input text.
        Args:
            text (str): Input text.
        Returns:
            str: Predicted emotion label.
        """
        inputs = self.tokenizer(text, return_tensors="pt", truncation=True, max_length=256)
        with torch.no_grad():
            outputs = self.model(**inputs)
            predicted_class = outputs.logits.argmax(dim=-1).item()
        # Handle cases where prediction index is not in label_map
        if predicted_class not in self.label_map:
            return f"Label_{predicted_class}"  # Return generic label if not found
        return self.label_map[predicted_class]
    
    def __call__(self, text):   
        return self.predict(text)


import os

# OpenAI
OPENAI_MODEL = "gpt-4o"
MAX_TOKENS = 500
TEMPERATURE = 0.7

# Project directories
PROJECT_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__))) # ./visonorm/visonorm
ARGS_PATH = os.path.join(PROJECT_PATH, 'arguments.json')
DATASET_DIR = os.path.join(PROJECT_PATH, "dataset")
CKPT_DIR = os.path.join(os.path.dirname(PROJECT_PATH), "model_checkpoints")
LOG_DIR = os.path.join(PROJECT_PATH, "logs")
DICT_PATH = os.path.join(PROJECT_PATH, "dictionary", "dictionary.json")
GIT_DOWNLOAD_URL = "https://github.com/AnhHoang0529/visonorm/releases/download/v0.0.1"

# Tokenizer constants
MASK_TOKEN = '<mask>'
BOS_TOKEN = '<s>'
EOS_TOKEN = '</s>'
PAD_TOKEN = '<pad>'
NULL_STR = '<space>'
UNK_TOKEN = '<unk>'
NUM_LABELS_N_MASKS = 5
NULL_STR_TO_SHOW = '_'
SPECIAL_TOKEN_LS = [NULL_STR, MASK_TOKEN, BOS_TOKEN, EOS_TOKEN, UNK_TOKEN]
PRETRAINED_TOKENIZER_MAP = {
    'phobert': 'vinai/phobert-base',
    'visobert': 'uitnlp/visobert',
    'bartpho': 'vinai/bartpho-syllable',
}

# Student constants
SAMPLES_PER_TASK_TO_REPORT = {"normalize": ["all", "NEED_NORM", "NORMED", "PRED_NEED_NORM", "PRED_NORMED"],
                              "n_masks_pred": ["all", "n_masks_1", "n_masks_2", "n_masks_3", "n_masks_4", "n_masks_5"]}
AVAILABLE_FINE_TUNING_STRATEGY = ["standard", "flexible_lr"]
RM_ACCENTS_DICT = str.maketrans(
    "ÁÀẢÃẠĂẮẰẲẴẶÂẤẦẨẪẬĐÈÉẺẼẸÊẾỀỂỄỆÍÌỈĨỊÓÒỎÕỌÔỐỒỔỖỘƠỚỜỞỠỢÚÙỦŨỤƯỨỪỬỮỰÝỲỶỸỴáàảãạăắằẳẵặâấầẩẫậđèéẻẽẹêếềểễệíìỉĩịóòỏõọôốồổỗộơớờởỡợúùủũụưứừửữựýỳỷỹỵ",
    "A"*17 + "D" + "E"*11 + "I"*5 + "O"*17 + "U"*11 + "Y"*5 + "a"*17 + "d" + "e"*11 + "i"*5 + "o"*17 + "u"*11 + "y"*5
)

ARGS = {
    "student_name": "bartpho",
    "max_token_len": 1024,
    "teacher_name": "ran",
    "training_mode": "weakly_supervised",
    "inference_model": "student",
    "ckpt_dir": CKPT_DIR,
    "logdir": LOG_DIR,
    "metric": "f1_score",
    "num_iter": 10,
    "num_rules": 2,
    "num_epochs": 10,
    "num_unsup_epochs": 5,
    "debug": 0,
    "remove_accents": 1,
    "rm_accent_ratio": 1.0,
    "append_n_mask": 1,
    "nsw_detect": 1,
    "soft_labels": 1,
    "loss_weights": 0,
    "hard_student_rule": 1,
    "train_batch_size": 16,
    "eval_batch_size": 128,
    "unsup_batch_size": 128,
    "lowercase": 1,
    "learning_rate": 0.001,
    "fine_tuning_strategy": "flexible_lr",
    "sample_size": 8096,
    "topk": 1,
    "seed": 42,
    "downstream_task": "vihsd"
}


import logging
from transformers import AutoTokenizer, AutoModelForMaskedLM
from .detect import NswDetector

# Configure logger once in your main script
logging.basicConfig(
    level=logging.INFO, 
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
LOGGER = logging.getLogger("visonorm.normalizer")

class ViSoLexNormalizer:
    def __init__(self, model_repo=None, device='cpu'):
        """
        Initialize the ViSoNorm Normalizer using HuggingFace model.
        
        Args:
            model_repo (str, optional): HuggingFace model repository. 
                If None, uses default model "visolex/visobert-normalizer-mix100"
            device (str): Device to run inference on ('cpu' or 'cuda')
        """
        if model_repo is None:
            model_repo = "visolex/visobert-normalizer-mix100"
        
        self.model_repo = model_repo
        self.device = device
        self.logger = LOGGER
        self.loaded = False
        
        # Load tokenizer and model
        self.logger.info(f"Loading tokenizer and model from {model_repo}...")
        self.tokenizer = AutoTokenizer.from_pretrained(model_repo)
        self.model = AutoModelForMaskedLM.from_pretrained(model_repo, trust_remote_code=True)
        self.model.to(device)
        self.loaded = True
        self.logger.info("✅ Model loaded successfully!")

    def normalize_sentence(self, input_str, detect_nsw=False):
        """
        Normalize a sentence and optionally detect NSW tokens.
        
        Args:
            input_str (str): Input text to normalize
            detect_nsw (bool): If True, also return NSW detection results
        
        Returns:
            If detect_nsw=False: normalized text string
            If detect_nsw=True: tuple of (nsw_spans, normalized_text)
                where nsw_spans is a list of dictionaries with NSW information
        """
        if not self.loaded:
            self.logger.warning("Model not loaded yet!")
            return input_str if not detect_nsw else ([], input_str)
        
        # Use the built-in normalize_text method from the model
        normalized_text, source_tokens, predicted_tokens = self.model.normalize_text(
            self.tokenizer, input_str, device=self.device
        )
        
        if detect_nsw:
            # Use the built-in detect_nsw method from the model
            nsw_results = self.model.detect_nsw(self.tokenizer, input_str, device=self.device)
            return nsw_results, normalized_text
        else:
            return normalized_text



"""
Optimization utilities for ViSoNorm training.
Includes improved optimizer, learning rate scheduling, and gradient clipping.
"""

import torch
import math
from typing import List, Optional, Union, Dict, Any
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler


def get_optimizer(parameters: List[torch.nn.Parameter], 
                  lr: float, 
                  optimizer_type: str = "adamw",
                  weight_decay: float = 0.01,
                  betas: tuple = (0.9, 0.999),
                  eps: float = 1e-8) -> Optimizer:
    """
    Create optimizer with improved settings.
    
    Args:
        parameters: Model parameters to optimize
        lr: Learning rate
        optimizer_type: Type of optimizer ("adam" or "adamw")
        weight_decay: Weight decay for regularization
        betas: Adam optimizer betas
        eps: Adam optimizer epsilon
        
    Returns:
        Configured optimizer
    """
    if optimizer_type.lower() == "adamw":
        return torch.optim.AdamW(
            parameters, 
            lr=lr, 
            betas=betas, 
            eps=eps, 
            weight_decay=weight_decay
        )
    elif optimizer_type.lower() == "adam":
        return torch.optim.Adam(
            parameters, 
            lr=lr, 
            betas=betas, 
            eps=eps, 
            weight_decay=weight_decay
        )
    else:
        raise ValueError(f"Unsupported optimizer type: {optimizer_type}")


def create_learning_rate_scheduler(optimizer: Optimizer,
                                 scheduler_type: str = "cosine",
                                 total_epochs: int = 10,
                                 warmup_epochs: int = 2,
                                 min_lr_ratio: float = 0.1,
                                 **kwargs) -> Optional[_LRScheduler]:
    """
    Create learning rate scheduler with warmup and decay.
    
    Args:
        optimizer: Optimizer to schedule
        scheduler_type: Type of scheduler ("cosine", "linear", "plateau", "none")
        total_epochs: Total number of training epochs
        warmup_epochs: Number of warmup epochs
        min_lr_ratio: Minimum learning rate ratio (min_lr = base_lr * ratio)
        **kwargs: Additional scheduler-specific arguments
        
    Returns:
        Configured scheduler or None
    """
    if scheduler_type.lower() == "none":
        return None
    
    # Validate parameters
    if total_epochs <= 0:
        raise ValueError(f"total_epochs must be positive, got {total_epochs}")
    if warmup_epochs < 0:
        raise ValueError(f"warmup_epochs must be non-negative, got {warmup_epochs}")
    if warmup_epochs >= total_epochs:
        print(f"Warning: warmup_epochs ({warmup_epochs}) >= total_epochs ({total_epochs}). "
              f"Consider reducing warmup_epochs or increasing total_epochs.")
    
    base_lr = optimizer.param_groups[0]['lr']
    min_lr = base_lr * min_lr_ratio
    
    if scheduler_type.lower() == "cosine":
        # Ensure T_0 is at least 1
        T_0 = max(1, total_epochs - warmup_epochs)
        return torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer,
            T_0=T_0,
            T_mult=1,
            eta_min=min_lr,
            last_epoch=-1
        )
    
    elif scheduler_type.lower() == "linear":
        def lr_lambda(epoch):
            if epoch < warmup_epochs:
                return epoch / max(1, warmup_epochs)  # Avoid division by zero
            else:
                # Handle case where warmup_epochs >= total_epochs
                if warmup_epochs >= total_epochs:
                    return 1.0
                progress = (epoch - warmup_epochs) / (total_epochs - warmup_epochs)
                return max(min_lr_ratio, 1.0 - progress)
        
        return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)
    
    elif scheduler_type.lower() == "plateau":
        return torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            mode='min',
            factor=0.5,
            patience=kwargs.get('patience', 2),
            verbose=True,
            min_lr=min_lr
        )
    
    else:
        raise ValueError(f"Unsupported scheduler type: {scheduler_type}")


def apply_fine_tuning_strategy_improved(fine_tuning_strategy: str, 
                                      model: torch.nn.Module, 
                                      lr_init: Union[float, Dict[str, float]],
                                      config: Any,
                                      append_n_mask: bool = False, 
                                      nsw_detector: bool = False) -> tuple:
    """
    Apply improved fine-tuning strategy with better optimizer and scheduler.
    
    Args:
        fine_tuning_strategy: Strategy type ("standard" or "flexible_lr")
        model: Model to optimize
        lr_init: Initial learning rate(s)
        config: Configuration object with optimizer and scheduler settings
        append_n_mask: Whether model has mask predictor
        nsw_detector: Whether model has NSW detector
        
    Returns:
        Tuple of (optimizer, scheduler)
    """
    from config import AVAILABLE_FINE_TUNING_STRATEGY
    
    assert fine_tuning_strategy in AVAILABLE_FINE_TUNING_STRATEGY, \
        f"{fine_tuning_strategy} not in {AVAILABLE_FINE_TUNING_STRATEGY}"
    
    if fine_tuning_strategy == "standard":
        assert isinstance(lr_init, float), f"lr_init should be float for {fine_tuning_strategy}"
        
        optimizer = get_optimizer(
            model.parameters(),
            lr=lr_init,
            optimizer_type=config.model.optimizer,
            weight_decay=config.model.weight_decay,
            betas=config.model.adam_betas,
            eps=config.model.adam_eps
        )
        
        scheduler = create_learning_rate_scheduler(
            optimizer,
            scheduler_type=config.model.scheduler_type if config.model.use_scheduler else "none",
            total_epochs=config.training.num_epochs,
            warmup_epochs=min(config.model.warmup_epochs, config.training.num_epochs - 1),
            min_lr_ratio=config.model.min_lr_ratio
        )
        
        print(f"TRAINING: fine tuning strategy {fine_tuning_strategy} : "
              f"optimizer {config.model.optimizer}, lr {lr_init}, "
              f"weight_decay {config.model.weight_decay}, scheduler {config.model.scheduler_type}")
        
        return [optimizer], scheduler
    
    else:  # fine_tuning_strategy == "flexible_lr"
        # Decide LRs based on the actual HuggingFace model variant (base vs large)
        # Priority: use model name from config; fallback to hidden size
        base_model = getattr(getattr(config, 'model', None), 'base_model', None)
        if hasattr(base_model, 'value'):
            base_model = base_model.value

        # Resolve HF model name provided on CLI for the selected architecture
        hf_model_name = getattr(config.model, 'pretrained_model_name', None)

        # Detect model size by explicit mapping of known variants first, then fallbacks
        is_large_by_name = False
        if isinstance(hf_model_name, str):
            name_l = hf_model_name.lower()
            if base_model == 'phobert':
                if 'phobert-large' in name_l:
                    is_large_by_name = True
                elif ('phobert-base' in name_l) or ('phobert-base-v2' in name_l):
                    is_large_by_name = False
            elif base_model == 'bartpho':
                # bartpho-syllable is large; bartpho-syllable-base is base
                if 'bartpho-syllable-base' in name_l:
                    is_large_by_name = False
                elif 'bartpho-syllable' in name_l:
                    is_large_by_name = True
            elif base_model == 'visobert':
                # only one version; treat as base
                is_large_by_name = False
            else:
                # generic fallback: contains 'large'
                is_large_by_name = ('large' in name_l)
        # Hidden size from the instantiated model
        hidden_size = None
        try:
            if hasattr(model, 'roberta'):
                hidden_size = model.roberta.embeddings.word_embeddings.weight.size(1)
            elif hasattr(model, 'bart'):
                hidden_size = model.bart.shared.weight.size(1)
            elif hasattr(model, 't5'):
                # T5 encoder-only backbone (vit5)
                hidden_size = model.t5.encoder.embed_tokens.weight.size(1)
        except Exception:
            hidden_size = None

        is_large_by_dim = (hidden_size is not None and hidden_size >= 1000)
        is_large = is_large_by_name or is_large_by_dim

        # Choose conservative LRs for large models
        if is_large:
            lr_emb = 5e-6
            lr_encoder = 1e-5
            lr_head = 1e-5
        else:
            lr_emb = 1e-5
            lr_encoder = 2e-5
            lr_head = 2e-5

        # Define learning rates for different model components by architecture
        if hasattr(model, 'roberta'):  # PhoBERT / ViSoBERT
            lr_dict = {
                "roberta.embeddings": lr_emb,
                "roberta.encoder": lr_encoder,
                "roberta.pooler": lr_head,
                "cls": lr_head,
            }
        elif hasattr(model, 'bart'):  # BARTpho
            lr_dict = {
                "bart.shared": lr_emb,
                "bart.encoder": lr_encoder,
                "bart.decoder": lr_encoder,
                "cls": lr_head,
            }
        elif hasattr(model, 't5'):  # ViT5 (T5 encoder-only)
            # T5EncoderModel ties embeddings to t5.shared; include final layer norm
            lr_dict = {
                "t5.shared": lr_emb,
                "t5.encoder.block": lr_encoder,
                "t5.encoder.final_layer_norm": lr_head,
                "cls": lr_head,
            }
        else:
            # Fallback: apply single head rate to all params
            lr_dict = {"": lr_head}
        
        if append_n_mask:
            lr_dict["mask_n_predictor"] = lr_head
        if nsw_detector:
            lr_dict["nsw_detector"] = lr_head
        
        # Create parameter groups
        param_groups = []
        n_all_layers = len([a for a, _ in model.named_parameters()])
        n_optim_layer = 0
        
        for prefix, lr in lr_dict.items():
            param_group = [param for name, param in model.named_parameters() if name.startswith(prefix)] if prefix != "" else [p for _, p in model.named_parameters()]
            if param_group:  # Only add if parameters exist
                param_groups.append({
                    'params': param_group,
                    'lr': lr
                })
                n_optim_layer += len(param_group)
        
        # Check if all parameters are covered
        if n_all_layers != n_optim_layer:
            print(f"WARNING: Not all layers covered in optimization. "
                  f"All layers: {n_all_layers}, Optimized layers: {n_optim_layer}")
        
        # Create optimizer for each parameter group
        optimizers = []
        for param_group in param_groups:
            optimizer = get_optimizer(
                param_group['params'],
                lr=param_group['lr'],
                optimizer_type=config.model.optimizer,
                weight_decay=config.model.weight_decay,
                betas=config.model.adam_betas,
                eps=config.model.adam_eps
            )
            optimizers.append(optimizer)
        
        # Create scheduler for the first optimizer (main model)
        scheduler = create_learning_rate_scheduler(
            optimizers[0],
            scheduler_type=config.model.scheduler_type if config.model.use_scheduler else "none",
            total_epochs=config.training.num_epochs,
            warmup_epochs=min(config.model.warmup_epochs, config.training.num_epochs - 1),
            min_lr_ratio=config.model.min_lr_ratio
        )
        
        print(f"TRAINING: fine tuning strategy {fine_tuning_strategy} : "
              f"optimizer {config.model.optimizer}, lr_dict {lr_dict}, "
              f"weight_decay {config.model.weight_decay}, scheduler {config.model.scheduler_type}")
        
        return optimizers, scheduler


def clip_gradients(model: torch.nn.Module, max_grad_norm: float = 1.0) -> float:
    """
    Clip gradients to prevent exploding gradients.
    
    Args:
        model: Model whose gradients to clip
        max_grad_norm: Maximum gradient norm
        
    Returns:
        Gradient norm before clipping
    """
    return torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=max_grad_norm)


def step_optimizers(optimizers: List[Optimizer], 
                   scheduler: Optional[_LRScheduler] = None,
                   model: Optional[torch.nn.Module] = None,
                   max_grad_norm: float = 1.0) -> Dict[str, float]:
    """
    Step optimizers with gradient clipping and scheduler update.
    
    Args:
        optimizers: List of optimizers to step
        scheduler: Learning rate scheduler (optional)
        model: Model for gradient clipping (optional)
        max_grad_norm: Maximum gradient norm for clipping
        
    Returns:
        Dictionary with gradient norm and learning rates
    """
    results = {}
    
    # Clip gradients if model is provided
    if model is not None:
        grad_norm = clip_gradients(model, max_grad_norm)
        results['grad_norm'] = grad_norm.item()
    
    # Step all optimizers
    for i, optimizer in enumerate(optimizers):
        optimizer.step()
        optimizer.zero_grad()
        
        # Get current learning rate
        current_lr = optimizer.param_groups[0]['lr']
        results[f'optimizer_{i}_lr'] = current_lr
    
    # Step scheduler if provided
    if scheduler is not None:
        if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            # For ReduceLROnPlateau, we need to pass the loss value
            # This should be called separately with the loss value
            pass
        else:
            scheduler.step()
            results['scheduler_lr'] = scheduler.get_last_lr()[0]
    
    return results



"""
Binary predictor for NSW (Non-Standard Word) detection.
"""

import torch.nn as nn


class BinaryPredictor(nn.Module):
    """
    Binary predictor for detecting non-standard words (NSW).
    
    Args:
        input_dim: Input dimension from encoder
        dense_dim: Hidden dimension for first dense layer (optional)
        dense_dim_2: Hidden dimension for second dense layer (optional)
        verbose: Verbosity level
    """
    
    def __init__(self, input_dim, dense_dim=None, dense_dim_2=None, verbose=1):
        super(BinaryPredictor, self).__init__()
        self.verbose = verbose
        self.dense_2 = None
        self.dense = None
        dim_predictor = input_dim
        
        if dense_dim is not None:
            if dense_dim > 0:
                self.dense = nn.Linear(input_dim, dense_dim)
                dim_predictor = dense_dim
                if dense_dim_2 is not None:
                    if dense_dim_2 > 0:
                        self.dense_2 = nn.Linear(dense_dim, dense_dim_2)
                        dim_predictor = dense_dim_2
        else:
            assert dense_dim_2 is None or dense_dim_2 == 0, "ERROR: dense_dim_2 cannot be not null if dense_dim is None"
        
        self.predictor = nn.Linear(dim_predictor, out_features=2)

    def forward(self, encoder_state_projected):
        """
        Forward pass for binary prediction.
        
        Args:
            encoder_state_projected: Encoder output [batch, seq_len, hidden_dim]
            
        Returns:
            Binary predictions [batch, seq_len, 2]
        """
        if self.dense is not None:
            intermediary = nn.ReLU()(self.dense(encoder_state_projected))
            if self.dense_2 is not None:
                intermediary = nn.ReLU()(self.dense_2(intermediary))
        else:
            intermediary = encoder_state_projected
        
        # Binary classification output [batch, seq_len, 2]
        hidden_state_normalize_not = self.predictor(intermediary)
        
        return hidden_state_normalize_not



DEFAULT_NSW_SEARCH_PROMPT = """
- Hãy chuẩn hóa từ {nsw} thành dạng chuẩn

- Giải nghĩa từ {nsw} với từng dạng chuẩn

- Thêm một số biến thể viết tắt cho mỗi dạng chuẩn của {nsw}  (nếu có)

- Đặt một câu ví dụ với từ không chuẩn {nsw}.

Kết quả trả về ở dạng HTML.
"""

DEFAULT_LEXNORM_PROMPT = """examples"""


from .downloader import DatasetDownloader
from .registry import DatasetRegistry

def list_datasets():
    """
    List all available datasets.
    Returns:
        list: List of dataset names.
    """
    return DatasetRegistry.list_datasets()

def load_dataset(name: str, force_download: bool = False):
    """
    Load a dataset by name.
    Args:
        name (str): Dataset name.
        force_download (bool): Force re-download even if cached.
    Returns:
        pd.DataFrame: Loaded dataset.
    """
    return DatasetDownloader.load_dataset(name, force_download)

def get_dataset_info(name: str):
    """
    Get metadata for a dataset by name.
    Args:
        name (str): Dataset name.
    Returns:
        dict: Metadata for the dataset.
    """
    return DatasetRegistry.get_dataset_info(name)

__all__ = ['list_datasets', 'load_dataset', 'get_dataset_info']


import torch
from transformers import AutoTokenizer, AutoModel
from contextlib import nullcontext
from visonorm.utils import suppress_output, clear_console

class AspectSentimentAnalysis:
    MODEL_MAP = {
        "smartphone": {
            "phobert-v1": "visolex/phobert-absa-smartphone",
            "phobert-v2": "visolex/phobert-v2-absa-smartphone",
            "bartpho": "visolex/bartpho-absa-smartphone",
            "vit5": "visolex/vit5-absa-smartphone",
            "mbert": "visolex/mbert-absa-smartphone",
            "visobert": "visolex/visobert-absa-smartphone",
            #"bilstm": "visolex/bilstm-absa-smartphone",
            "xlm-roberta": "visolex/xlm-roberta-absa-smartphone",
            #"textcnn": "visolex/textcnn-absa-smartphone"
        },
        "restaurant": {
            "phobert-v1": "visolex/phobert-v1-absa-restaurant",
            "phobert-v2": "visolex/phobert-v2-absa-restaurant",            
            "bartpho": "visolex/bartpho-absa-restaurant",
            #"bilstm": "visolex/bilstm-absa-restaurant",
            "vit5": "visolex/vit5-absa-restaurant",
            "mbert": "visolex/mbert-absa-restaurant",
            "visobert": "visolex/visobert-absa-restaurant",
            "xlm-roberta": "visolex/xlm-roberta-absa-restaurant",
            #"textcnn": "visolex/textcnn-absa-restaurant"
        },
        "hotel": {
            "phobert-v1": "visolex/phobert-v1-absa-hotel",
            "phobert-v2": "visolex/phobert-v2-absa-hotel",            
            "bartpho": "visolex/bartpho-absa-hotel",
            "vit5": "visolex/vit5-absa-hotel",
            "mbert": "visolex/mbert-absa-hotel",
            "visobert": "visolex/visobert-absa-hotel",
            #"bilstm": "visolex/bilstm-absa-hotel",
            "xlm-roberta": "visolex/xlm-roberta-absa-hotel",
            #"textcnn": "visolex/textcnn-absa-hotel"
        }
    }

    DOMAIN_ASPECT_LABELS = {
        "smartphone": [
            "BATTERY", "CAMERA", "DESIGN", "FEATURES", "GENERAL",
            "PERFORMANCE", "PRICE", "SCREEN", "SERandACC", "STORAGE"
        ],
        "hotel": [
            "FACILITIES#CLEANLINESS", "FACILITIES#COMFORT", "FACILITIES#DESIGN&FEATURES",
            "FACILITIES#GENERAL", "FACILITIES#MISCELLANEOUS", "FACILITIES#PRICES",
            "FACILITIES#QUALITY", "FOOD&DRINKS#MISCELLANEOUS", "FOOD&DRINKS#PRICES",
            "FOOD&DRINKS#QUALITY", "FOOD&DRINKS#STYLE&OPTIONS", "HOTEL#CLEANLINESS",
            "HOTEL#COMFORT", "HOTEL#DESIGN&FEATURES", "HOTEL#GENERAL",
            "HOTEL#MISCELLANEOUS", "HOTEL#PRICES", "HOTEL#QUALITY", "LOCATION#GENERAL",
            "ROOMS#CLEANLINESS", "ROOMS#COMFORT", "ROOMS#DESIGN&FEATURES",
            "ROOMS#GENERAL", "ROOMS#MISCELLANEOUS", "ROOMS#PRICES", "ROOMS#QUALITY",
            "ROOM_AMENITIES#CLEANLINESS", "ROOM_AMENITIES#COMFORT",
            "ROOM_AMENITIES#DESIGN&FEATURES", "ROOM_AMENITIES#GENERAL",
            "ROOM_AMENITIES#MISCELLANEOUS", "ROOM_AMENITIES#PRICES",
            "ROOM_AMENITIES#QUALITY", "SERVICE#GENERAL"
        ],
        "restaurant": [
            "AMBIENCE#GENERAL", "DRINKS#PRICES", "DRINKS#QUALITY", "DRINKS#STYLE&OPTIONS",
            "FOOD#PRICES", "FOOD#QUALITY", "FOOD#STYLE&OPTIONS", "LOCATION#GENERAL",
            "RESTAURANT#GENERAL", "RESTAURANT#MISCELLANEOUS", "RESTAURANT#PRICES",
            "SERVICE#GENERAL"
        ]
    }

    @staticmethod
    def list_domains():
        """
        List all available domains for aspect sentiment analysis.
        Returns:
            list: List of available domain names.
        """
        return list(AspectSentimentAnalysis.MODEL_MAP.keys())
    
    @staticmethod
    def list_models(domain=None):
        """
        List all available models for aspect sentiment analysis.
        Args:
            domain (str, optional): Domain name. If None, returns all models for all domains.
        Returns:
            dict: Dictionary with domain names as keys and model dictionaries as values,
                  or if domain is specified, returns model dictionary for that domain.
        """
        if domain is None:
            return AspectSentimentAnalysis.MODEL_MAP.copy()
        if domain not in AspectSentimentAnalysis.MODEL_MAP:
            available = AspectSentimentAnalysis.list_domains()
            raise ValueError(f"Domain '{domain}' not found. Available domains: {available}")
        return AspectSentimentAnalysis.MODEL_MAP[domain].copy()
    
    @staticmethod
    def list_model_names(domain):
        """
        List all available model names for a specific domain.
        Args:
            domain (str): Domain name.
        Returns:
            list: List of available model names for the specified domain.
        """
        if domain not in AspectSentimentAnalysis.MODEL_MAP:
            available = AspectSentimentAnalysis.list_domains()
            raise ValueError(f"Domain '{domain}' not found. Available domains: {available}")
        return list(AspectSentimentAnalysis.MODEL_MAP[domain].keys())

    def __init__(self, domain, model_name="phobert", sentiment_labels=["POSITIVE", "NEGATIVE", "NEUTRAL"], silent=True, clear_output=True):
        if domain not in self.MODEL_MAP:
            available = self.list_domains()
            raise ValueError(f"Domain '{domain}' not found. Available domains: {available}")
        if model_name not in self.MODEL_MAP[domain]:
            available = self.list_model_names(domain)
            raise ValueError(f"Model '{model_name}' not found for domain '{domain}'. Available models: {available}")
        model_path = self.MODEL_MAP[domain][model_name]
        ctx = suppress_output() if silent else nullcontext()
        with ctx:
            self.tokenizer = AutoTokenizer.from_pretrained(model_path, trust_remote_code=True)
            self.model = AutoModel.from_pretrained(model_path, trust_remote_code=True)
        if clear_output:
            clear_console()
        self.aspect_labels = self.DOMAIN_ASPECT_LABELS.get(domain, [])
        self.sentiment_labels = sentiment_labels

    def predict(self, text, threshold=0.5):
        """
        Perform aspect-based sentiment analysis on the input text.
        Args:
            text (str): Input text.
            threshold (float): Confidence threshold for predictions.
        Returns:
            list: List of tuples (aspect, sentiment).
        """
        inputs = self.tokenizer(text, return_tensors="pt", padding=True, truncation=True, max_length=256)
        inputs.pop("token_type_ids", None)
        with torch.no_grad():
            out = self.model(**inputs)
        logits = out.logits.squeeze(0)
        probs = torch.softmax(logits, dim=-1)
        none_id = probs.size(-1) - 1
        results = []
        for i, aspect in enumerate(self.aspect_labels):
            prob_i = probs[i]
            pred_id = int(prob_i.argmax().item())
            if pred_id != none_id and pred_id < len(self.sentiment_labels):
                score = prob_i[pred_id].item()
                if score >= threshold:
                    results.append((aspect, self.sentiment_labels[pred_id].lower()))
        return results
    
    def __call__(self, text, domain, model_name="phobert", threshold=0.5):
        """
        Call the predict method with the specified domain and model name.
        Args:
            text (str): Input text.
            domain (str): Domain to use for aspect-based sentiment analysis.
            model_name (str): Model name to use.
            threshold (float): Confidence threshold for predictions.
        Returns:
            list: List of tuples (aspect, sentiment).
        """
        return self.predict(text, threshold)


import math
from collections import OrderedDict
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import CrossEntropyLoss
from transformers import RobertaConfig, AutoModel
from config import NUM_LABELS_N_MASKS
from .binary_predictor import BinaryPredictor

def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

try:
    # apex.normalization.fused_layer_norm.FusedLayerNorm is optimized for better performance on GPU architectures compared to nn.LayerNorm
    from apex.normalization.fused_layer_norm import FusedLayerNorm as RobertaLayerNorm
except ImportError:
    print("Better speed can be achieved with apex installed from https://www.github.com/nvidia/apex.")
    class RobertaLayerNorm(nn.Module):
        def __init__(self, hidden_size, eps=1e-12):
            """Construct a layernorm module in the TF style (epsilon inside the square root).
            """
            super(RobertaLayerNorm, self).__init__()
            self.weight = nn.Parameter(torch.ones(hidden_size))
            self.bias = nn.Parameter(torch.zeros(hidden_size))
            self.variance_epsilon = eps

        def forward(self, x):
            u = x.mean(-1, keepdim=True)
            s = (x - u).pow(2).mean(-1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.variance_epsilon)
            return self.weight * x + self.bias

class RobertaLMHead(nn.Module):
    def __init__(self, config, roberta_model_embedding_weights):
        super().__init__()
        # Use the actual hidden size from the pretrained model, not the config
        actual_hidden_size = roberta_model_embedding_weights.size(1)
        self.dense = nn.Linear(actual_hidden_size, actual_hidden_size)
        self.layer_norm = RobertaLayerNorm(actual_hidden_size, eps=1e-12)

        num_labels = roberta_model_embedding_weights.size(0)
        self.decoder = nn.Linear(actual_hidden_size, num_labels, bias=False)
        self.decoder.weight = roberta_model_embedding_weights
        self.decoder.bias = nn.Parameter(torch.zeros(num_labels))

    def forward(self, features):
        x = self.dense(features)
        x = gelu(x)
        x = self.layer_norm(x)
        x = self.decoder(x)
        return x

class RobertaMaskNPredictionHead(nn.Module):
    def __init__(self, config, actual_hidden_size):
        super(RobertaMaskNPredictionHead, self).__init__()
        self.mask_predictor_dense = nn.Linear(actual_hidden_size, 50)
        self.mask_predictor_proj = nn.Linear(50, NUM_LABELS_N_MASKS)
        self.activation = gelu

    def forward(self, sequence_output):
        mask_predictor_state = self.activation(self.mask_predictor_dense(sequence_output))
        prediction_scores = self.mask_predictor_proj(mask_predictor_state)
        return prediction_scores

class RobertaForMaskedLM(nn.Module):
    def __init__(self, vocab_size, mask_n_predictor, nsw_detector, model_name='vinai/phobert-base'):
        super(RobertaForMaskedLM, self).__init__()
        self.roberta = AutoModel.from_pretrained(model_name)
        self.roberta.resize_token_embeddings(vocab_size)
        self.roberta.config.vocab_size = vocab_size
        self.roberta.config.mask_n_predictor = mask_n_predictor
        self.roberta.config.nsw_detector = nsw_detector
        self.config = self.roberta.config
        # Get actual hidden size from the pretrained model
        self.actual_hidden_size = self.roberta.embeddings.word_embeddings.weight.size(1)
        
        self.cls = RobertaLMHead(self.config, self.roberta.embeddings.word_embeddings.weight)
        self.mask_n_predictor = RobertaMaskNPredictionHead(self.config, self.actual_hidden_size) if self.config.mask_n_predictor else None
        self.nsw_detector = BinaryPredictor(self.actual_hidden_size, dense_dim=100) if self.config.nsw_detector else None
        self.num_labels_n_mask = NUM_LABELS_N_MASKS

    def forward(self, input_ids, token_type_ids=None, attention_mask=None,
                labels=None, labels_n_masks=None, standard_labels=None, sample_weights=None,
                soft_labels=False):
        # Ensure masks exist and are correct dtype
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)
        pad_id = getattr(self.roberta.config, 'pad_token_id', None)
        if pad_id is None:
            pad_id = getattr(self.roberta.embeddings.word_embeddings, 'padding_idx', 1)
        if attention_mask is None:
            attention_mask = (input_ids != pad_id).long()
        else:
            attention_mask = attention_mask.long()

        # Final safety: truncate to smallest supported positional length
        cfg_max = int(getattr(self.roberta.config, 'max_position_embeddings', input_ids.size(1)))
        tbl_max = int(getattr(self.roberta.embeddings.position_embeddings, 'num_embeddings', cfg_max))
        max_pos = min(cfg_max, tbl_max)
        # Reserve 2 positions for special tokens (<s>, </s>) per PhoBERT convention
        eff_max = max(1, max_pos - 2)
        if input_ids.size(1) > eff_max:
            input_ids = input_ids[:, :eff_max]
            token_type_ids = token_type_ids[:, :eff_max]
            attention_mask = attention_mask[:, :eff_max]

        # Build clamped position_ids from attention_mask to avoid OOB indices
        position_ids = attention_mask.cumsum(dim=1) - 1
        position_ids.masked_fill_(attention_mask == 0, 0)
        position_ids.clamp_(min=0, max=max_pos - 1)

        outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            output_hidden_states=True,
        )
        sequence_output = outputs.last_hidden_state

        # Calculate predictions
        prediction_scores = self.cls(sequence_output)  # (batch_size, seq_len, vocab_size)

        loss_dict = OrderedDict([("loss", None), ("loss_norm", 0), ("loss_n_masks_pred", 0), ("loss_nsw_detection", 0)])
        pred_dict = OrderedDict([("logits_norm", None), ("logits_n_masks_pred", None), ("logits_nsw_detection", None)])

        if self.mask_n_predictor is not None:
            assert self.num_labels_n_mask > 0, "ERROR  "
            logits_n_mask_prediction = self.mask_n_predictor(sequence_output)
            pred_dict["logits_n_masks_pred"] = logits_n_mask_prediction
        
        # Calculate predictions for NSW detection
        if self.nsw_detector is not None:
            standard_logits = self.nsw_detector(sequence_output)
            pred_dict["logits_nsw_detection"] = standard_logits
        
        pred_dict["logits_norm"] = prediction_scores

        # Calculate loss
        if labels is not None:
            if self.mask_n_predictor is not None:
                assert labels_n_masks is not None, "ERROR : you provided labels for normalization and self.mask_n_predictor : so you should provide labels_n_mask_prediction"
                if sample_weights is None:
                    loss_fct_masks_pred = CrossEntropyLoss(ignore_index=-1)
                    loss_dict["loss_n_masks_pred"] = loss_fct_masks_pred(logits_n_mask_prediction.view(-1, self.num_labels_n_mask), labels_n_masks.view(-1))
                else:
                    loss_fct_masks_pred = CrossEntropyLoss(ignore_index=-1, reduce='none')
                    loss_dict["loss_n_masks_pred"] = (loss_fct_masks_pred(logits_n_mask_prediction.view(-1, self.num_labels_n_mask), labels_n_masks.view(-1)) * sample_weights).mean()

            if self.nsw_detector is not None:
                assert standard_labels is not None, "ERROR : you provided labels for normalization and self.nsw_detector : so you should provide standard_labels"
                if sample_weights is None:
                    loss_fct_nsw_pred = CrossEntropyLoss(ignore_index=-1)
                    loss_dict["loss_nsw_detection"] = loss_fct_nsw_pred(standard_logits.view(-1, 2), standard_labels.view(-1))
                else:
                    loss_fct_nsw_pred = CrossEntropyLoss(ignore_index=-1, reduce='none')
                    loss_dict["loss_nsw_detection"] = (loss_fct_nsw_pred(standard_logits.view(-1, 2), standard_labels.view(-1)) * sample_weights).mean()

            # Use the actual vocabulary size from the model (not vocab_size + 1)
            num_labels = self.config.vocab_size
            masked_lm_labels = labels.view(-1, num_labels) if soft_labels else labels.view(-1)

            if sample_weights is None:
                loss_fct = CrossEntropyLoss(ignore_index=-1)
                masked_lm_loss = loss_fct(prediction_scores.view(-1, num_labels), masked_lm_labels)
            else:
                loss_fct = CrossEntropyLoss(ignore_index=-1, reduce='none')
                masked_lm_loss = (loss_fct(prediction_scores.view(-1, num_labels), masked_lm_labels)*sample_weights).mean()
            loss_dict["loss_norm"] = masked_lm_loss

        loss_dict["loss"] = loss_dict["loss_norm"] + loss_dict["loss_n_masks_pred"] + loss_dict["loss_nsw_detection"]

        return loss_dict, pred_dict, sequence_output


def get_phobert_normalizer(vocab_size, checkpoint_dir=None, mask_n_predictor=False, nsw_detector=False, model_name='vinai/phobert-base'):
    """
    Create a PhoBERT model for text normalization.
    
    Args:
        vocab_size: Size of the vocabulary
        checkpoint_dir: Path to checkpoint file (optional)
        mask_n_predictor: Whether to include N-mask prediction head
        nsw_detector: Whether to include NSW detection head
        model_name: HuggingFace model name
        
    Returns:
        RobertaForMaskedLM model instance
    """
    model = RobertaForMaskedLM(
        vocab_size=vocab_size,
        mask_n_predictor=mask_n_predictor,
        nsw_detector=nsw_detector,
        model_name=model_name
    )
    
    # The model.cls.decoder is already properly initialized in RobertaForMaskedLM.__init__
    # We only need to handle checkpoint loading if provided
    if checkpoint_dir is not None:
        model.load_state_dict(torch.load(checkpoint_dir, map_location=lambda storage, loc: storage))

    return model



import re
from itertools import chain
from typing import List, Dict, Any
from .regex_expression import (
    EdgePunctLeft, EdgePunctRight, Protected, Emoji_Protected, emoji_pattern, emoji_list, tone_dict_map
)

# Utility: Tone normalization
def tone_normalization(text: str, tone_map: Dict[str, str] = tone_dict_map) -> str:
    """
    Replace text based on a tone mapping dictionary.
    
    Args:
        text (str): Input text to be normalized.
        tone_map (dict): Mapping dictionary with replacement rules.
    
    Returns:
        str: Normalized text.
    """
    for original, replacement in tone_map.items():
        text = text.replace(replacement, original)
    return text

# Utility: Remove redundant dots not at the end of the list
def remove_dot_not_end(lst: List[str]) -> List[str]:
    """
    Remove dots ('.') from the list except if at the end or followed by an uppercase letter.
    
    Args:
        lst (list): Input list of strings.
    
    Returns:
        list: Modified list with redundant dots removed.
    """
    for i, elem in enumerate(lst):
        if elem == '.':
            if i != len(lst) - 1 and (i == 0 or not (lst[i + 1][0].isupper() if lst[i + 1] and len(lst[i + 1]) > 0 else False)):
                lst[i] = ''
    return lst

# Utility: Flatten nested lists
def flatten(lst: List[Any]) -> List[Any]:
    """
    Recursively flatten a nested list.
    
    Args:
        lst (list): Input list (potentially nested).
    
    Returns:
        list: Flattened list.
    """
    return list(chain.from_iterable(flatten(item) if isinstance(item, list) else [item] for item in lst))

# Tokenization helpers
def split_edge_punctuation(text: str, edge_punct_left: re.Pattern = EdgePunctLeft, edge_punct_right: re.Pattern = EdgePunctRight) -> str:
    """
    Split text by edge punctuation rules.
    
    Args:
        text (str): Input text.
        edge_punct_left (Pattern): Regex for punctuation on the left.
        edge_punct_right (Pattern): Regex for punctuation on the right.
    
    Returns:
        str: Modified text.
    """
    text = edge_punct_left.sub(r"\1\2 \3", text)
    text = edge_punct_right.sub(r"\1 \2\3", text)
    return text

def add_non_empty(target: List[str], source: List[str]) -> List[str]:
    """
    Add non-empty strings from one list to another.
    
    Args:
        target (list): List to append to.
        source (list): List with potential elements to append.
    
    Returns:
        list: Combined list with non-empty elements added.
    """
    for item in source:
        stripped_item = item.strip()
        if stripped_item:
            target.append(stripped_item)
    return target

# Emoji handling
def split_emoji_text(text: str, emoji_list: List[str]) -> str:
    """
    Separate emojis from text.
    
    Args:
        text (str): Input text.
        emoji_list (list): List of emoji characters.
    
    Returns:
        str: Text with emojis separated and extra whitespace removed.
    """
    # Split text and add spaces around emojis
    result = []
    for char in text:
        if char in emoji_list:
            result.append(f' {char} ')
        else:
            result.append(char)
    
    # Join and remove extra whitespace
    joined_text = ''.join(result)
    # Remove multiple consecutive spaces and trim leading/trailing whitespace
    cleaned_text = re.sub(r'\s+', ' ', joined_text).strip()
    return cleaned_text

def split_emoji_emoji(text_array: List[str], emoji_list: List[str]) -> List[str]:
    """
    Split emojis that are grouped together in words.
    
    Args:
        text_array (list): List of strings or words.
        emoji_list (list): List of emoji characters.
    
    Returns:
        list: Modified list with separated emojis.
    """
    result = []
    for word in text_array:
        if any(char in emoji_list for char in word):
            for char in word:
                result.append(char if char in emoji_list else word)
        else:
            result.append(word)
    return ' '.join(result)


# Main tokenization logic
def simple_tokenize(text: str, protected_patterns: List[re.Pattern], emoji_pattern: re.Pattern) -> List[str]:
    """
    Tokenize text while protecting specific patterns and emojis.
    
    Args:
        text (str): Input text to tokenize.
        protected_patterns (list): List of regex patterns to protect.
        emoji_pattern (Pattern): Regex pattern for emojis.
    
    Returns:
        list: Tokenized text as a list of strings.
    """
    split_punct_text = split_edge_punctuation(text)

    # Detect spans of protected patterns and emojis
    bad_spans = []
    bads = []
    for pattern in protected_patterns:
        for match in pattern.finditer(split_punct_text):
            if match.start() != match.end():
                bads.append([split_punct_text[match.start():match.end()]])
                bad_spans.append((match.start(), match.end()))

    # Find indices for "good" (non-protected) spans
    indices = [0]
    for start, end in bad_spans:
        indices.extend([start, end])
    indices.append(len(split_punct_text))

    # Extract good and bad spans
    split_goods = [
        split_punct_text[indices[i]:indices[i + 1]].strip().split()
        for i in range(0, len(indices), 2)
    ]

    # Interleave good and bad spans
    result = []
    for i, bad in enumerate(bads):
        result = add_non_empty(result, split_goods[i])
        result = add_non_empty(result, bad)
    result = add_non_empty(result, split_goods[-1])
    return result

# Case folding utility
def case_folding(text, mode="lower"):
        """
        Perform case folding on the input text.
        Args:
            text (str): Input text.
            mode (str): "lower", "upper", or "capitalize".
        Returns:
            str: Case-folded text.
        """
        if mode == "lower":
            return text.lower()
        elif mode == "upper":
            return text.upper()
        elif mode == "capitalize":
            return text.title()
        else:
            raise ValueError("Invalid mode. Choose 'lower', 'upper', or 'capitalize'.")



def detect_emoji(text: str) -> List[str]:
    """
    Detect emojis in the text while respecting protected patterns.
    
    Args:
        text (str): Input text to check for emojis.
 
    
    Returns:
        list: List of detected emojis.
    """
    splitPunctText = split_edge_punctuation(text)

    textLength = len(splitPunctText)

    bads = []
    badSpans = []
    for match in Emoji_Protected.finditer(splitPunctText):
        # The spans of the "bads" should not be split.
        if (match.start() != match.end()): #unnecessary?
            bads.append( [splitPunctText[match.start():match.end()]] )
            badSpans.append( (match.start(), match.end()) )

    indices = [0]
    for start, end in badSpans:
        indices.extend([start, end])
    indices.append(textLength)

    result = []

    for i, bad in enumerate(bads):
        result = add_non_empty(result, bad)
    
    return result

def remove_emojis(text: str) -> str:
    """
    Remove emojis from the text.
    
    Args:
        text (str): Input text from which emojis will be removed.
    
    Returns:
        str: Text with emojis removed.
    """
    splitPunctText = split_edge_punctuation(text)

    textLength = len(splitPunctText)

    bads = []
    badSpans = []

    for match in Emoji_Protected.finditer(splitPunctText):
        # The spans of the "bads" should not be split.
        if (match.start() != match.end()):
            bads.append([splitPunctText[match.start():match.end()]])
            badSpans.append((match.start(), match.end()))
    indices = [0]
    for start, end in badSpans:
        indices.extend([start, end])
    indices.append(textLength)

    split_goods = [
        splitPunctText[indices[i]:indices[i + 1]].strip().split()
        for i in range(0, len(indices), 2)
    ]

    result = []

    for i, good in enumerate(split_goods):
        result = add_non_empty(result, good)

    return ' '.join(result)




#!/usr/bin/env python3
"""
Setup script for ViSoNorm package
"""

from setuptools import setup, find_packages
import os

# Read the README file
def read_readme():
    with open("README.md", "r", encoding="utf-8") as fh:
        return fh.read()

# Read the requirements file
def read_requirements():
    with open("requirements.txt", "r", encoding="utf-8") as fh:
        return [line.strip() for line in fh if line.strip() and not line.startswith("#")]

# Get version from VERSION file
def get_version():
    version_file = os.path.join("visonorm", "VERSION")
    if os.path.exists(version_file):
        with open(version_file, "r") as f:
            return f.read().strip()
    return "0.0.1"

setup(
    name="visonorm",
    version=get_version(),
    author="Ha Dung Nguyen",
    author_email="dungngh@uit.edu.vn",
    description="Vietnamese Social Media Lexical Normalization Toolkit",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/AnhHoang0529/visonorm",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.10",
    install_requires=read_requirements(),
    extras_require={
        "dev": [
            "pytest>=6.0",
            "black>=22.0",
            "flake8>=4.0",
        ],
    },
    include_package_data=True,
    package_data={
        "visonorm": ["VERSION", "*.json", "*.txt"],
    },
    zip_safe=False,
) 


#!/usr/bin/env python3
"""
Script to upload trained ViSoNorm student model to Hugging Face Hub.
Supports ViSoBERT, PhoBERT, and BARTpho models.
"""

import os
import json
import argparse
import torch
from pathlib import Path
from typing import Optional, Dict, Any
from transformers import AutoTokenizer, AutoConfig
from huggingface_hub import HfApi, Repository, create_repo, hf_hub_download
import shutil

from config import Config, BaseModel, PRETRAINED_TOKENIZER_MAP
from base_model.Student import Student
from base_model.ViSoNormalizer import ViSoNormalizerTrainer


def analyze_checkpoint_weights(state_dict, base_model: str) -> dict:
    """Analyze presence of expected ViSoNorm head weights vs common alternatives.

    Returns a report dictionary with counts and example keys to help detect issues
    like missing heads or mismatched parameter names.
    """
    keys = list(state_dict.keys())

    expected_heads = [
        'cls_decoder.weight', 'cls_decoder.bias',
        'cls_dense.weight', 'cls_dense.bias',
        'cls_layer_norm.weight', 'cls_layer_norm.bias',
        'mask_n_predictor.mask_predictor_dense.weight', 'mask_n_predictor.mask_predictor_dense.bias',
        'mask_n_predictor.mask_predictor_proj.weight', 'mask_n_predictor.mask_predictor_proj.bias',
        'nsw_detector.dense.weight', 'nsw_detector.dense.bias',
        'nsw_detector.predictor.weight', 'nsw_detector.predictor.bias',
    ]

    alt_common_heads = [
        'lm_head.weight', 'lm_head.bias',
        'cls.decoder.weight', 'cls.decoder.bias',
        'cls.dense.weight', 'cls.dense.bias',
        'cls.layer_norm.weight', 'cls.layer_norm.bias',
    ]

    aux_heads = [
        'nsw_detector.', 'mask_n_predictor.',
    ]

    def present(name):
        return any(k == name or k.startswith(name) for k in keys)

    report = {
        'base_model': base_model,
        'total_params': len(keys),
        'expected_heads_present': {name: present(name) for name in expected_heads},
        'alt_common_heads_present': {name: present(name) for name in alt_common_heads},
        'aux_heads_present': {n: any(n in k for k in keys) for n in aux_heads},
        'example_keys': keys[:20],
    }

    return report


def create_model_card(model_name: str, base_model: str, performance: Dict[str, Any], 
                     training_config: Dict[str, Any]) -> str:
    """Create a model card for the uploaded model."""
    
    model_card = f"""---
    license: mit
    tags:
    - text-normalization
    - vietnamese
    - lexical-normalization
    - visonorm
    - {base_model}
    pipeline_tag: fill-mask
    ---

    # {model_name}

    This model is a Vietnamese text normalization model trained using the ViSoNorm framework with {base_model.upper()} architecture.

    ## Model Description

    This model performs lexical normalization for Vietnamese text, converting informal text to standard Vietnamese. It was trained using the ViSoNorm (Self-training with Weak Supervision) framework.

    ## Performance

    """

    if performance:
        model_card += "| Metric | Score |\n"
        model_card += "|--------|-------|\n"
        for metric, score in performance.items():
            if isinstance(score, (int, float)):
                model_card += f"| {metric} | {score:.2f}% |\n"

    model_card += f"""
        ## Training Configuration

        - **Base Model**: {base_model.upper()}
        - **Training Mode**: {training_config.get('training_mode', 'N/A')}
        - **Learning Rate**: {training_config.get('learning_rate', 'N/A')}
        - **Epochs**: {training_config.get('num_epochs', 'N/A')}
        - **Batch Size**: {training_config.get('train_batch_size', 'N/A')}

        ## Usage

        ```python
        from transformers import AutoTokenizer, AutoModelForMaskedLM

        # Load model and tokenizer
        model_repo = "your-username/your-model-name"  # Replace with your actual repo
        tokenizer = AutoTokenizer.from_pretrained(model_repo)
        model = AutoModelForMaskedLM.from_pretrained(model_repo, trust_remote_code=True)

        # Normalize text using the built-in method
        text = "sv dh gia dinh chua cho di lam :))"
        normalized_text, source_tokens, predicted_tokens = model.normalize_text(
            tokenizer, text, device='cpu'
        )

        # Output: sinh viên đại học gia đình chưa cho đi làm :))
        ```

        ## Example Outputs

        | Input | Output |
        |-------|--------|
        | `sv dh gia dinh chua cho di lam :))` | `sinh viên đại học gia đình chưa cho đi làm :))` |
        | `chúng nó bảo em là ctrai` | `chúng nó bảo em là con trai` |
        | `anh ơi em muốn đi chơi` | `anh ơi em muốn đi chơi` |

        ## Citation

        If you use this model, please cite the ViSoNorm paper:

        ```bibtex
        @article{{visonorm2024,
        title={{ViSoNorm: Self-training with Weak Supervision for Vietnamese Text Normalization}},
        author={{Your Name}},
        journal={{arXiv preprint}},
        year={{2024}}
        }}
        ```
        """
    
    return model_card


def get_model_config_from_state_dict(model_path):
    """Extract configuration parameters from the actual model file"""
    try:
        # Handle both file and directory paths
        if os.path.isfile(model_path):
            # Direct file path provided
            state_dict_path = model_path
        else:
            # Directory path provided, look for model files inside
            model_file_bin = os.path.join(model_path, "pytorch_model.bin")
            model_file_pt = os.path.join(model_path, "final_model.pt")
            if os.path.exists(model_file_bin):
                state_dict_path = model_file_bin
            elif os.path.exists(model_file_pt):
                state_dict_path = model_file_pt
            else:
                print(f"Warning: No model file found in directory {model_path}")
                return {}
        
        # Load the model state dict
        state_dict = torch.load(state_dict_path, map_location='cpu')
        
        config_info = {}
        
        
        # Look for specific embedding keys based on the actual model structure
        for key, tensor in state_dict.items():
            # ViSoBERT/PhoBERT embeddings
            if key == 'roberta.embeddings.word_embeddings.weight':
                vocab_size = tensor.shape[0]
                config_info['vocab_size'] = vocab_size
            elif key == 'roberta.embeddings.position_embeddings.weight':
                max_pos = tensor.shape[0]
                config_info['max_position_embeddings'] = max_pos
            elif key == 'roberta.embeddings.token_type_embeddings.weight':
                type_vocab_size = tensor.shape[0]
                config_info['type_vocab_size'] = type_vocab_size
            # BartPho embeddings
            elif key == 'model.shared.weight':
                vocab_size = tensor.shape[0]
                config_info['vocab_size'] = vocab_size
            elif key == 'model.encoder.embed_positions.weight':
                max_pos = tensor.shape[0]
                config_info['max_position_embeddings'] = max_pos
            elif key == 'model.decoder.embed_positions.weight':
                max_pos = tensor.shape[0]
                config_info['max_position_embeddings'] = max_pos
            # Alternative BartPho embedding keys
            elif 'shared.weight' in key and 'vocab_size' not in config_info:
                vocab_size = tensor.shape[0]
                config_info['vocab_size'] = vocab_size
            elif 'embed_positions.weight' in key and 'max_position_embeddings' not in config_info:
                max_pos = tensor.shape[0]
                config_info['max_position_embeddings'] = max_pos
            elif 'positional_embedding' in key.lower() and 'max_position_embeddings' not in config_info:
                max_pos = tensor.shape[0]
                config_info['max_position_embeddings'] = max_pos
        
        # Look for hidden size from any embedding
        for key, tensor in state_dict.items():
            if 'embedding' in key.lower() and 'weight' in key.lower() and len(tensor.shape) == 2:
                hidden_size = tensor.shape[1]
                config_info['hidden_size'] = hidden_size
                break
        
        # Fallback: if we still don't have vocab_size, try to find it from any word embedding
        if 'vocab_size' not in config_info:
            for key, tensor in state_dict.items():
                if ('word' in key.lower() or 'shared' in key.lower()) and 'weight' in key.lower() and len(tensor.shape) == 2:
                    vocab_size = tensor.shape[0]
                    config_info['vocab_size'] = vocab_size
                    break
        
        # Fallback: if we still don't have max_position_embeddings, try to find it from any position embedding
        if 'max_position_embeddings' not in config_info:
            for key, tensor in state_dict.items():
                if ('position' in key.lower() or 'pos' in key.lower()) and 'weight' in key.lower() and len(tensor.shape) == 2:
                    max_pos = tensor.shape[0]
                    config_info['max_position_embeddings'] = max_pos
                    break
        
        # Look for number of layers from transformer blocks
        layer_count = 0
        for key in state_dict.keys():
            if 'layer.' in key and 'weight' in key:
                try:
                    layer_num = int(key.split('layer.')[1].split('.')[0])
                    layer_count = max(layer_count, layer_num + 1)
                except:
                    pass
        
        if layer_count > 0:
            config_info['num_hidden_layers'] = layer_count
        
        return config_info
        
    except Exception as e:
        print(f"Error reading model file: {e}")
        return {}


def assert_heads_present(model_bin_path: str) -> dict:
    """Return a presence report for required heads and raise if missing."""
    state_dict = torch.load(model_bin_path, map_location="cpu")
    def present(prefix: str) -> int:
        return sum(1 for k in state_dict.keys() if k.startswith(prefix))
    report = {
        "cls": present("cls."),
        "mask_n_predictor": present("mask_n_predictor."),
        "nsw_detector": present("nsw_detector."),
    }
    # Require both auxiliary heads to be present for multitask inference
    if report["mask_n_predictor"] == 0 or report["nsw_detector"] == 0:
        raise RuntimeError(
            "Missing head weights in checkpoint: "
            f"mask_n_predictor={report['mask_n_predictor']}, "
            f"nsw_detector={report['nsw_detector']}. "
            "Re-export from training ensuring all heads are saved."
        )
    return report


def prepare_model_for_upload(model_path: str, base_model: str, config: Config, hf_config_json: Optional[str] = None) -> Dict[str, Any]:
    """Prepare the model files for upload to Hugging Face."""
    
    # Create temporary directory for upload (clean up any existing one)
    temp_dir = Path(f"temp_upload_{base_model}")
    if temp_dir.exists():
        shutil.rmtree(temp_dir)
    temp_dir.mkdir(exist_ok=True)
    
    try:
        # Copy model files (prefer pytorch_model.bin, fallback to final_model.pt)
        # Check if model_path is a file or directory
        if os.path.isfile(model_path):
            # Direct file path provided
            shutil.copy2(model_path, temp_dir / "pytorch_model.bin")
        else:
            # Directory path provided, look for model files inside
            model_file_bin = os.path.join(model_path, "pytorch_model.bin")
            model_file_pt = os.path.join(model_path, "final_model.pt")
            if os.path.exists(model_file_bin):
                shutil.copy2(model_file_bin, temp_dir / "pytorch_model.bin")
            elif os.path.exists(model_file_pt):
                shutil.copy2(model_file_pt, temp_dir / "pytorch_model.bin")
            else:
                raise FileNotFoundError(f"Model file not found: {model_file_bin} or {model_file_pt}")

        # Enforce presence of multitask heads
        head_report = assert_heads_present(str(temp_dir / "pytorch_model.bin"))
        # Persist a small JSON report for transparency
        with open(temp_dir / "state_dict_report.json", "w", encoding="utf-8") as f:
            json.dump(head_report, f, indent=2)
        
        # Create or reuse config.json (prefer provided or existing; else download base and minimally override)
        model_type_mapping = {
            'visobert': 'xlm-roberta',
            'phobert': 'roberta', 
            'bartpho': 'mbart',  # BARTpho is based on mBART architecture
            'vit5': 't5',        # ViT5 uses T5 encoder architecture
        }
        architecture_mapping = {
            'visobert': 'ViSoNormViSoBERTForMaskedLM',
            'phobert': 'RobertaForMaskedLM',
            'bartpho': 'ViSoNormBartPhoForMaskedLM',  # Use our custom class
            'vit5': 'ViSoNormViT5ForMaskedLM',        # Custom ViT5 class
        }
        standard_model_type = model_type_mapping.get(base_model.lower(), 'roberta')
        standard_architecture = architecture_mapping.get(base_model.lower(), 'RobertaForMaskedLM')
        # If explicit HF config is provided, use it
        used_existing_config = False
        if hf_config_json and os.path.exists(hf_config_json):
            shutil.copy2(hf_config_json, temp_dir / "config.json")
            used_existing_config = True
        else:
            # Try to reuse config.json from model directory
            model_dir_for_config = model_path if os.path.isdir(model_path) else os.path.dirname(model_path)
            existing_cfg = os.path.join(model_dir_for_config, "config.json")
            if os.path.exists(existing_cfg):
                shutil.copy2(existing_cfg, temp_dir / "config.json")
                used_existing_config = True

        if not used_existing_config:
            # Download base config.json from the pretrained backbone and minimally override
            pretrained_name = getattr(getattr(config, 'model', None), 'pretrained_model_name', None)
            if not pretrained_name:
                raise ValueError("pretrained_model_name must be provided in Config.model.pretrained_model_name to download base config.json")
            try:
                base_cfg_path = hf_hub_download(repo_id=pretrained_name, filename="config.json")
                with open(base_cfg_path, "r", encoding="utf-8") as f:
                    base_cfg = json.load(f)
            except Exception as e:
                print(f"⚠️ Failed to download base config.json from {pretrained_name}: {e}. Falling back to checkpoint-derived config.")
                base_cfg = {}

            # Derive minimal necessary values from checkpoint/tokenizer
            model_config_info = get_model_config_from_state_dict(model_path)
            vocab_size = model_config_info.get('vocab_size', len(config.tokenizer) if hasattr(config, 'tokenizer') else base_cfg.get('vocab_size', 32000))

            # Merge overrides
            updated_cfg = dict(base_cfg)
            updated_cfg['vocab_size'] = int(vocab_size)
            updated_cfg['mask_n_predictor'] = True
            updated_cfg['nsw_detector'] = True
            # Ensure architectures/model_type minimally consistent if missing
            updated_cfg.setdefault('model_type', standard_model_type)
            updated_cfg.setdefault('architectures', [standard_architecture])

            with open(temp_dir / "config.json", "w", encoding="utf-8") as f:
                json.dump(updated_cfg, f, indent=2)

        # Ensure auto_map and architectures always point to our custom classes, even when reusing existing config
        try:
            with open(temp_dir / "config.json", "r", encoding="utf-8") as f:
                cfg_current = json.load(f)
        except Exception:
            cfg_current = {}

        auto_map = (
            {
                "AutoModel": "visonorm_visobert_model.ViSoNormViSoBERTForMaskedLM",
                "AutoModelForMaskedLM": "visonorm_visobert_model.ViSoNormViSoBERTForMaskedLM"
            } if base_model.lower() == 'visobert' else
            {
                "AutoModel": "visonorm_phobert_model.ViSoNormPhoBERTForMaskedLM",
                "AutoModelForMaskedLM": "visonorm_phobert_model.ViSoNormPhoBERTForMaskedLM"
            } if base_model.lower() == 'phobert' else
            {
                "AutoModel": "visonorm_bartpho_model.ViSoNormBartPhoForMaskedLM",
                "AutoModelForMaskedLM": "visonorm_bartpho_model.ViSoNormBartPhoForMaskedLM"
            } if base_model.lower() == 'bartpho' else
            {
                "AutoModel": "visonorm_vit5_model.ViSoNormViT5ForMaskedLM",
                "AutoModelForMaskedLM": "visonorm_vit5_model.ViSoNormViT5ForMaskedLM"
            } if base_model.lower() == 'vit5' else None
        )
        if auto_map is not None:
            cfg_current['auto_map'] = auto_map
            cfg_current['architectures'] = [standard_architecture]
            with open(temp_dir / "config.json", "w", encoding="utf-8") as f:
                json.dump(cfg_current, f, indent=2)
        
        # Create tokenizer config for better compatibility
        if base_model.lower() == 'bartpho':
            # BartPho requires specific tokenizer configuration
            tokenizer_config = {
                "tokenizer_class": "BartPhoTokenizer",
                "model_max_length": 512,
                "padding_side": "right",
                "truncation_side": "right",
                "pad_token": "<pad>",
                "bos_token": "<s>",
                "eos_token": "</s>",
                "unk_token": "<unk>",
                "mask_token": "<mask>",
                "additional_special_tokens": ["<pad>", "<s>", "</s>", "<unk>", "<mask>"],
                "use_fast": False,  # BartPho typically uses slow tokenizer
            }
        else:
            # Base tokenizer config - special tokens may vary by model
            additional_tokens = ["<pad>", "<s>", "</s>", "<unk>", "<mask>", "<space>"]
            
            tokenizer_config = {
                "tokenizer_class": "XLMRobertaTokenizer" if base_model.lower() == 'visobert' else 
                                  "RobertaTokenizer" if base_model.lower() == 'phobert' else 
                                  ("T5Tokenizer" if base_model.lower() == 'vit5' else "BartPhoTokenizer"),
                "model_max_length": 512,
                "padding_side": "right",
                "truncation_side": "right",
                "pad_token": "<pad>",
                "bos_token": "<s>",
                "eos_token": "</s>",
                "unk_token": "<unk>",
                "mask_token": "<mask>",
                "additional_special_tokens": additional_tokens
            }
        
        with open(temp_dir / "tokenizer_config.json", "w") as f:
            json.dump(tokenizer_config, f, indent=2)
        
        # Copy ViSoNorm custom model file based on architecture
        def get_architecture_model_file(base_model: str) -> str:
            """Get the appropriate ViSoNorm model file for the architecture."""
            base_model_lower = base_model.lower()
            if 'bartpho' in base_model_lower or 'bart' in base_model_lower:
                return "visonorm_bartpho_model.py"
            elif 'phobert' in base_model_lower or 'roberta' in base_model_lower:
                return "visonorm_phobert_model.py"
            elif 'visobert' in base_model_lower or 'xlm' in base_model_lower:
                return "visonorm_visobert_model.py"
            elif 'vit5' in base_model_lower or 't5' in base_model_lower:
                return "visonorm_vit5_model.py"
            else:
                # Default to ViSoBERT for unknown models
                return "visonorm_visobert_model.py"
        
        visonorm_model_file = get_architecture_model_file(base_model)
        if os.path.exists(visonorm_model_file):
            shutil.copy2(visonorm_model_file, temp_dir / visonorm_model_file)
        else:
            print(f"Warning: ViSoNorm model file not found: {visonorm_model_file}")
        
        # Copy README.md if it exists
        readme_file = "README.md"
        if os.path.exists(readme_file):
            shutil.copy2(readme_file, temp_dir / "README.md")
        
        # Determine model directory for tokenizer file copying
        if os.path.isfile(model_path):
            # If model_path is a file, look for tokenizer files in the parent directory
            model_dir = os.path.dirname(model_path)
        else:
            # If model_path is a directory, use it directly
            model_dir = model_path
        
        # For BartPho, we need to ensure we have the correct tokenizer files
        if base_model.lower() == 'bartpho':
            # CRITICAL: The training process modifies the tokenizer by adding special tokens like <space>
            # We MUST use the modified tokenizer files from training, not the original BartPho files
            
            # Priority 1: Use modified tokenizer files from training (these contain the <space> token)
            modified_tokenizer_files = [
                "tokenizer.json",           # Contains the full modified vocabulary
                "tokenizer_config.json",    # Contains tokenizer configuration
                "special_tokens_map.json",  # Contains special token mappings including <space>
                "added_tokens.json"        # Contains the added tokens (like <space>)
            ]
            
            # CRITICAL: Copy the BARTpho tokenization file for proper tokenizer behavior
            bartpho_tokenization_files = [
                "tokenization_bartpho.py",  # Custom BARTpho tokenizer implementation
                "bartpho_tokenization.py"   # Alternative naming
            ]
            
            tokenization_file_found = False
            for file_name in bartpho_tokenization_files:
                src_path = os.path.join(model_dir, file_name)
                if os.path.exists(src_path):
                    shutil.copy2(src_path, temp_dir / file_name)
                    tokenization_file_found = True
                    break
            
            if not tokenization_file_found:
                print("⚠️  BARTpho tokenization file not found! Attempting to download from vinai/bartpho-syllable...")
                
                try:
                    from huggingface_hub import hf_hub_download
                    # Download the tokenization file from the original BARTpho repository
                    tokenization_file = hf_hub_download(
                        repo_id="vinai/bartpho-syllable",
                        filename="tokenization_bartpho.py",
                        local_dir=temp_dir
                    )
                    tokenization_file_found = True
                except Exception as e:
                    print(f"❌ Failed to download BARTpho tokenization file: {e}. This may cause tokenization issues when loading the model.")
            
            modified_files_found = 0
            for file_name in modified_tokenizer_files:
                src_path = os.path.join(model_dir, file_name)
                if os.path.exists(src_path):
                    shutil.copy2(src_path, temp_dir / file_name)
                    modified_files_found += 1
                else:
                    # File missing; will rely on fallbacks below
                    pass
            # Priority 2: Copy original BartPho files as fallback (but these won't have <space> token)
            original_bartpho_files = [
                "vocab.txt",
                "monolingual_vocab.txt", 
                "dict.txt",
                "spm.model",
                "sentencepiece.bpe.model"
            ]
            
            for file_name in original_bartpho_files:
                src_path = os.path.join(model_dir, file_name)
                if os.path.exists(src_path):
                    shutil.copy2(src_path, temp_dir / file_name)
                else:
                    # File not present; continue
                    pass
            
            # Check if we have the critical modified files
            if modified_files_found < 2:
                print("🚨 CRITICAL WARNING: Missing modified tokenizer files! The model was trained with a modified tokenizer that includes the <space> token. Without these files, the model will predict weird characters.")
        
        # Copy tokenizer files if they exist
        tokenizer_path = os.path.join(model_dir, "tokenizer")
        if os.path.exists(tokenizer_path):
            shutil.copytree(tokenizer_path, temp_dir / "tokenizer", dirs_exist_ok=True)

        # Also copy common tokenizer artifacts from model_dir root if present
        maybe_tokenizer_files = [
            "tokenizer.json",
            "tokenizer_config.json",
            "special_tokens_map.json",
            "vocab.json",
            "merges.txt",
            "sentencepiece.model",
            "sentencepiece.bpe.model",
            # BartPho specific files
            "vocab.txt",
            "monolingual_vocab.txt",
            "dict.txt",
            "spm.model",
        ]
        
        copied_files = []
        for fname in maybe_tokenizer_files:
            src = os.path.join(model_dir, fname)
            if os.path.exists(src):
                shutil.copy2(src, temp_dir / fname)
                copied_files.append(fname)
        
        if not copied_files:
            print(f"Warning: No tokenizer files found in {model_dir}")
        
        # Create training arguments
        training_args = {
            "base_model": base_model,
            "training_mode": config.model.training_mode.value,
            "learning_rate": config.model.learning_rate,
            "num_epochs": config.training.num_epochs,
            "train_batch_size": config.training.train_batch_size,
            "eval_batch_size": config.training.eval_batch_size,
            "remove_accents": config.training.remove_accents,
            "lower_case": config.training.lower_case,
        }
        
        with open(temp_dir / "training_args.json", "w") as f:
            json.dump(training_args, f, indent=2)

        # Diagnostics: analyze checkpoint for expected heads vs alternatives
        try:
            if os.path.isfile(model_path):
                sd_path = model_path
            else:
                maybe_sd = os.path.join(model_path, "pytorch_model.bin")
                sd_path = maybe_sd if os.path.exists(maybe_sd) else None
            if sd_path:
                state_dict = torch.load(sd_path, map_location='cpu')
                diag = analyze_checkpoint_weights(state_dict, base_model)
                with open(temp_dir / "state_dict_report.json", "w") as f:
                    json.dump(diag, f, indent=2)
            else:
                print(f"Warning: No state dict found at {model_path}")
        except Exception as e:
            print(f"State dict analysis failed: {e}")
        
        # Read back the config.json we prepared for reporting
        try:
            with open(temp_dir / "config.json", "r", encoding="utf-8") as f:
                prepared_cfg = json.load(f)
        except Exception:
            prepared_cfg = {}
        return {
            "temp_dir": temp_dir,
            "model_config": prepared_cfg,
            "training_args": training_args
        }
        
    except Exception as e:
        # Cleanup on error
        if temp_dir.exists():
            shutil.rmtree(temp_dir)
        raise e


def upload_model_to_huggingface(
    model_path: str,
    base_model: str,
    repo_name: str,
    performance: Optional[Dict[str, Any]] = None,
    config: Optional[Config] = None,
    token: Optional[str] = None,
    hf_config_json: Optional[str] = None
):
    """Upload the trained model to Hugging Face Hub."""
    
    print(f"Preparing {base_model} model for upload...")
    
    # Prepare model files
    prep_result = prepare_model_for_upload(model_path, base_model, config, hf_config_json=hf_config_json)
    temp_dir = prep_result["temp_dir"]
    model_config = prep_result["model_config"]
    training_args = prep_result["training_args"]
    
    try:
        # Create repository
        print(f"Creating repository: {repo_name}")
        create_repo(repo_id=repo_name, token=token, exist_ok=True)
        
        # Initialize repository in a clean directory
        repo_dir = Path(f"repo_{base_model}")
        if repo_dir.exists():
            shutil.rmtree(repo_dir)
            
        repo = Repository(
            local_dir=str(repo_dir),
            clone_from=repo_name,
            token=token
        )
        
        # Copy our prepared files to the repository
        for file_path in temp_dir.glob("*"):
            if file_path.is_file():
                shutil.copy2(file_path, repo_dir / file_path.name)
            elif file_path.is_dir():
                shutil.copytree(file_path, repo_dir / file_path.name, dirs_exist_ok=True)
        
        # Create model card
        model_card = create_model_card(
            repo_name, 
            base_model, 
            performance or {}, 
            training_args
        )
        
        with open(temp_dir / "README.md", "w", encoding="utf-8") as f:
            f.write(model_card)
        
        # Upload files
        print("Uploading model files...")
        repo.push_to_hub(commit_message="Upload ViSoNorm trained model")
        
        print(f"✅ Successfully uploaded model to: https://huggingface.co/{repo_name}")
        
    except Exception as e:
        print(f"❌ Error uploading model: {str(e)}")
        raise e
    
    finally:
        # Cleanup temporary directories
        if temp_dir.exists():
            shutil.rmtree(temp_dir)
        repo_dir = Path(f"repo_{base_model}")
        if repo_dir.exists():
            shutil.rmtree(repo_dir)


def main():
    parser = argparse.ArgumentParser(description="Upload ViSoNorm model to Hugging Face Hub")
    parser.add_argument("--model_path", type=str, required=True, help="Path to the saved model directory (e.g., ./experiments/visobert/weakly_supervised/student_best)")
    parser.add_argument("--base_model", type=str, required=True, choices=['visobert', 'phobert', 'bartpho', 'vit5'], help="Name of the base model (backbone)")
    parser.add_argument("--repo_name", type=str, required=True, help="Hugging Face repository name (e.g., username/vietnamese-text-normalizer)")
    parser.add_argument("--performance_file", type=str, required=False, help="Path to JSON file containing model performance metrics")
    parser.add_argument("--config_file", type=str, required=False, help="Path to config file used for training")
    parser.add_argument("--token", type=str, required=False, help="Hugging Face API token (or set HF_TOKEN environment variable)")
    parser.add_argument("--hf_config_json", type=str, required=False, help="Optional path to an existing Hugging Face config.json to use for upload")
    
    args = parser.parse_args()
    
    # Load performance metrics if provided
    performance = {}
    if args.performance_file and os.path.exists(args.performance_file):
        with open(args.performance_file, "r") as f:
            performance = json.load(f)
    
    # Load config if provided
    config = None
    if args.config_file and os.path.exists(args.config_file):
        # Load config from file
        with open(args.config_file, "r") as f:
            config_dict = json.load(f)
        config = Config()
        # Update config with loaded values
        for key, value in config_dict.items():
            if hasattr(config, key):
                setattr(config, key, value)
    else:
        # Create default config
        config = Config()
        config.model.base_model = BaseModel(args.base_model)
    
    # Get token from environment if not provided
    token = args.token or os.getenv("HF_TOKEN")
    if not token:
        print("❌ Please provide Hugging Face token via --token argument or HF_TOKEN environment variable")
        return
    
    # Upload model
    upload_model_to_huggingface(
        model_path=args.model_path,
        base_model=args.base_model,
        repo_name=args.repo_name,
        performance=performance,
        config=config,
        token=token,
        hf_config_json=args.hf_config_json
    )


if __name__ == "__main__":
    main()



import logging
from transformers import AutoTokenizer, AutoModelForMaskedLM

# Configure logger once in your main script
logging.basicConfig(
    level=logging.INFO, 
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
LOGGER = logging.getLogger("visonorm.nsw_detector")

class NswDetector:
    def __init__(self, model_repo=None, device='cpu'):
        """
        Initialize the NSW Detector using HuggingFace model.
        
        Args:
            model_repo (str, optional): HuggingFace model repository. 
                If None, uses default model "visolex/visobert-normalizer-mix100"
            device (str): Device to run inference on ('cpu' or 'cuda')
        """
        if model_repo is None:
            model_repo = "visolex/visobert-normalizer-mix100"
        
        self.model_repo = model_repo
        self.device = device
        self.logger = LOGGER
        self.loaded = False
        
        # Load tokenizer and model
        self.logger.info(f"Loading tokenizer and model from {model_repo}...")
        self.tokenizer = AutoTokenizer.from_pretrained(model_repo)
        self.model = AutoModelForMaskedLM.from_pretrained(model_repo, trust_remote_code=True)
        self.model.to(device)
        self.loaded = True
        self.logger.info("✅ Model loaded successfully!")

    def detect_nsw(self, input_str):
        """
        Detect Non-Standard Words (NSW) in the input text.
        
        Args:
            input_str (str): Input text to analyze
        
        Returns:
            List of dictionaries containing NSW information:
            [{'index': int, 'start_index': int, 'end_index': int, 'nsw': str, 
              'prediction': str, 'confidence_score': float}, ...]
        """
        if not self.loaded:
            self.logger.warning("Model not loaded yet!")
            return []
        
        # Use the built-in detect_nsw method from the model
        nsw_results = self.model.detect_nsw(self.tokenizer, input_str, device=self.device)
        
        return nsw_results

    def concatenate_nsw_spans(self, nsw_spans):
        """
        Concatenate adjacent NSW spans.
        
        Args:
            nsw_spans: List of NSW span dictionaries
        
        Returns:
            List of concatenated NSW spans
        """
        if not nsw_spans:
            return []
        
        result = []
        current_span = nsw_spans[0].copy()

        for i in range(1, len(nsw_spans)):
            next_span = nsw_spans[i]
            # Check if spans are adjacent
            if 'end_index' in current_span and 'start_index' in next_span:
                if current_span['end_index'] == next_span['start_index']:
                    current_span['nsw'] += next_span['nsw']
                    current_span['end_index'] = next_span['end_index']
                    if 'prediction' in next_span:
                        current_span['prediction'] = current_span.get('prediction', '') + ' ' + next_span.get('prediction', '')
                    continue
            result.append(current_span)
            current_span = next_span.copy()
        result.append(current_span)

        return result



from typing import List
from visonorm.module.utils.regex_expression import Protected, Emoji_Protected, emoji_list, tone_dict_map
from visonorm.module.utils.text_preprocessing import split_emoji_text, split_emoji_emoji, detect_emoji, remove_emojis
class EmojiHandler:
    def __init__(self):
        """Initialize EmojiHandler with optional logging (if needed)."""
        self.logger = None  # Placeholder for logger if needed in the future.

    def split_emoji_text(self, input_str: str):
        """
        Split text by emojis and return a list of strings.
        
        Args:
            input_str (str): Input text to be split.
        
        Returns:
            str: Text with emojis separated.
        """
        return split_emoji_text(input_str, emoji_list=emoji_list)

    def split_emoji_emoji(self, input_str: str) -> List[str]:
        """
        Split consecutive emojis in the input string.
        
        Args:
            input_str (str): Input text containing emojis.
        
        Returns:
            List[str]: List of strings with consecutive emojis split.
        """
        return split_emoji_emoji(input_str.split(), emoji_list=emoji_list)

    def detect_emoji(self, input_str: str) -> List[str]:
        """
        Detect emojis in the input string.
        
        Args:
            input_str (str): Input text to detect emojis.
        
        Returns:
            List[str]: List of detected emojis.
        """
        return detect_emoji(input_str)
    
    def remove_emojis(self, input_str: str) -> str:
        """
        Remove emojis from the input string.
        
        Args:
            input_str (str): Input text from which emojis will be removed.
        
        Returns:
            str: Text with emojis removed.
        """
        return remove_emojis(input_str)


from .hate_speech_detection import HateSpeechDetection
from .hate_speech_span_detection import HateSpeechSpanDetection
from .spam_review_detection import SpamReviewDetection
from .emotion_recognition import EmotionRecognition
from .aspect_sentiment_analysis import AspectSentimentAnalysis

__all__ = [
    "HateSpeechDetection",
    "HateSpeechSpanDetection",
    "SpamReviewDetection",
    "EmotionRecognition",
    "AspectSentimentAnalysis"
]




"""
Training loop implementation for ViSoNorm.
Separates training logic from main orchestration.
"""
import logging
from typing import Dict, List, Any, Optional
import numpy as np

from config import Config
from base_model.Student import Student
from teacher.Teacher import Teacher
from core.data import DataHandler
from core.evaluation import Evaluator
from utils import evaluate_model as evaluate, write_predictions, save_and_report_results


class TrainingLoop:
    """Handles the main training loop for ViSoNorm."""
    
    def __init__(self, config: Config, logger: logging.Logger):
        self.config = config
        self.logger = logger
        self.evaluator = Evaluator(config, logger=logger)
        
        # Initialize components
        self.data_handler = None
        self.student = None
        self.teacher = None
        self.tokenizer = None
        
        # Results tracking
        self.results = {}
        self.performance_history = {
            'teacher_dev': [],
            'teacher_test': [],
            'teacher_train': [],
            'student_dev': [],
            'student_test': [],
            'student_train': []
        }
        self.student_predictions = []
    
    def initialize_components(self, tokenizer):
        """Initialize all training components."""
        self.tokenizer = tokenizer
        self.data_handler = DataHandler(self.config, tokenizer=tokenizer, logger=self.logger)
        self.student = Student(self.config, tokenizer=tokenizer, logger=self.logger)
        
        if self.config.model.training_mode.value == 'weakly_supervised':
            self.teacher = Teacher(self.config, tokenizer=tokenizer, logger=self.logger)
            self.teacher.student = self.student
            self._setup_teacher_dimensions()
    
    def _setup_teacher_dimensions(self):
        """Setup teacher model dimensions based on student model.

        Each backbone implementation exposes `actual_hidden_size`; prefer that.
        """
        model = self.student.trainer.model
        if hasattr(model, 'actual_hidden_size'):
            actual_hidden_size = model.actual_hidden_size
        else:
            # Safe fallbacks in case a model variant does not expose the attribute
            if hasattr(model, 'roberta'):
                actual_hidden_size = model.roberta.embeddings.word_embeddings.weight.size(1)
            elif hasattr(model, 'bart'):
                actual_hidden_size = model.bart.shared.weight.size(1)
            elif hasattr(model, 't5'):
                try:
                    actual_hidden_size = model.t5.encoder.embed_tokens.weight.size(1)
                except Exception:
                    actual_hidden_size = int(getattr(model.t5.config, 'd_model'))
            else:
                # Final fallback to config hidden size
                actual_hidden_size = int(getattr(getattr(model, 'config', object()), 'hidden_size', 768))
        
        self.teacher.agg_model.xdim = actual_hidden_size
        self.logger.info(f"Set RAN xdim to {actual_hidden_size} for {self.config.model.base_model.value}")
    
    def load_datasets(self):
        """Load all required datasets."""
        self.logger.info("Loading datasets")
        datasets = {}
        for split in ['train', 'dev', 'test', 'unlabeled']:
            datasets[split] = self.data_handler.load_dataset(method=split)
        
        # Create pseudo-dataset for unlabeled data
        datasets['pseudo'] = self.data_handler.create_pseudodataset(datasets['unlabeled'])
        datasets['pseudo'].downsample(self.config.training.sample_size)
        
        return datasets
    
    def train_initial_student(self, datasets: Dict[str, Any]) -> Dict[str, Any]:
        """Train student on initial labeled data."""
        self.logger.info("*** Training Student on labeled data ***")
        
        train_pseudo = self.data_handler.create_pseudodataset(datasets['train'])
        results = self.student.train(
            train_dataset=train_pseudo, 
            dev_dataset=datasets['dev'], 
            mode='train'
        )
        
        self.performance_history['student_train'].append(results)
        
        if self.config.model.training_mode.value == 'only_student':
            self.student.save('student_best')
        
        return results
    
    def evaluate_initial_student(self, datasets: Dict[str, Any]) -> Dict[str, Any]:
        """Evaluate student on dev and test sets."""
        # Dev evaluation
        self.logger.info("*** Evaluating student on dev data ***")
        dev_results = evaluate(
            self.student, datasets['dev'], self.evaluator,
            comment="student dev", 
            remove_accents=self.config.training.remove_accents
        )
        self.performance_history['student_dev'].append(dev_results)
        
        # Test evaluation
        self.logger.info("*** Evaluating student on test data ***")
        test_results, test_predictions = evaluate(
            self.student, datasets['test'], self.evaluator,
            "test", comment="student test",
            remove_accents=self.config.training.remove_accents
        )
        self.performance_history['student_test'].append(test_results)
        
        # Save predictions
        write_predictions(
            self.config, self.logger, self.tokenizer, 
            test_predictions, file_name="student_test_predictions"
        )
        
        return {
            'dev': dev_results,
            'test': test_results,
            'predictions': test_predictions
        }
    
    def initialize_teacher_results(self):
        """Initialize teacher results with zeros."""
        zero_results = {
            'precision': 0, 'recall': 0, 'f1_score': 0, 
            'accuracy': 0, 'perf': 0
        }
        self.performance_history['teacher_train'].append(zero_results)
        self.performance_history['teacher_dev'].append(zero_results)
        self.performance_history['teacher_test'].append(zero_results)
    
    def run_iteration(self, iteration: int, datasets: Dict[str, Any]) -> Dict[str, Any]:
        """Run a single training iteration."""
        self.logger.info(f"*** Starting loop {iteration+1}/{self.config.training.num_iter} ***")
        np.random.seed(self.config.system.seed + iteration)
        
        # Downsample pseudo-dataset
        datasets['pseudo'].downsample(self.config.training.sample_size)
        
        iteration_results = {}
        
        if self.config.model.training_mode.value == 'weakly_supervised':
            iteration_results.update(self._run_teacher_iteration(datasets, iteration))
        else:
            iteration_results.update(self._run_self_training_iteration(datasets, iteration))
        
        # Re-train student on pseudo-labeled data
        self.logger.info('Re-train student on pseudo-labeled instances')
        pseudo_train_results = self.student.train(
            train_dataset=datasets['pseudo'], 
            dev_dataset=datasets['dev'], 
            mode='train_pseudo'
        )
        
        # Fine-tune student on clean labeled data
        self.logger.info('Fine-tuning the student on clean labeled data')
        train_pseudo = self.data_handler.create_pseudodataset(datasets['train'])
        finetune_results = self.student.train(
            train_dataset=train_pseudo, 
            dev_dataset=datasets['dev'], 
            mode='finetune'
        )
        
        self.performance_history['student_train'].append(finetune_results)
        
        # Evaluate student
        iteration_results.update(self._evaluate_student_iteration(datasets, iteration))
        
        # Save best model if improved
        self._save_best_model_if_improved(iteration_results['dev'])
        
        # Periodic memory cleanup to avoid RAM growth across long runs
        self._periodic_cleanup()
        
        return iteration_results
    
    def _run_teacher_iteration(self, datasets: Dict[str, Any], iteration: int) -> Dict[str, Any]:
        """Run teacher-based iteration."""
        # Train teacher
        self.teacher.train_ran(
            train_dataset=datasets['train'],
            dev_dataset=datasets['dev'],
            unlabeled_dataset=datasets['pseudo']
        )
        
        # Apply teacher on unlabeled data
        teacher_predictions = self.teacher.predict_ran(dataset=datasets['pseudo'])
        
        # Evaluate teacher
        self.logger.info("*** Evaluating teacher on dev data ***")
        teacher_dev_results, _ = evaluate(
            self.teacher, datasets['dev'], self.evaluator,
            "ran", comment=f"teacher dev iter{iteration+1}"
        )
        self.performance_history['teacher_dev'].append(teacher_dev_results)
        
        self.logger.info("*** Evaluating teacher on test data ***")
        teacher_test_results, _ = evaluate(
            self.teacher, datasets['test'], self.evaluator,
            "ran", comment=f"teacher test iter{iteration+1}"
        )
        self.performance_history['teacher_test'].append(teacher_test_results)
        
        # Update pseudo-dataset with teacher predictions
        self._update_pseudo_dataset_with_teacher(datasets['pseudo'], teacher_predictions)
        
        return {
            'teacher_dev': teacher_dev_results,
            'teacher_test': teacher_test_results
        }
    
    def _run_self_training_iteration(self, datasets: Dict[str, Any]) -> Dict[str, Any]:
        """Run self-training iteration."""
        from utils import sort_data
        
        sorted_pseudo = sort_data(datasets['pseudo'])
        student_predictions = self.student.predict(dataset=sorted_pseudo)
        
        # Update pseudo-dataset with student predictions
        self._update_pseudo_dataset_with_student(datasets['pseudo'], student_predictions)
        
        return {}
    
    def _update_pseudo_dataset_with_teacher(self, pseudo_dataset, teacher_predictions):
        """Update pseudo-dataset with teacher predictions."""
        self.logger.info("Update unlabeled data with Teacher's predictions")
        pseudo_dataset.teacher_data['id'] = teacher_predictions['id']
        pseudo_dataset.teacher_data['input_ids'] = teacher_predictions['input_ids']
        pseudo_dataset.teacher_data['align_index'] = teacher_predictions['align_index']
        pseudo_dataset.teacher_data['labels'] = teacher_predictions['preds']
        # If soft labels are required later, keep proba (downcast to float16 to save RAM);
        # otherwise, compute lightweight weights and drop proba to save memory
        if 'proba' in teacher_predictions and teacher_predictions['proba'] is not None:
            if self.config.model.soft_labels:
                pseudo_dataset.teacher_data['proba'] = [
                    arr.astype(np.float16) for arr in teacher_predictions['proba']
                ]
                pseudo_dataset.teacher_data['weights'] = [
                    np.max(arr, axis=-1) for arr in teacher_predictions['proba']
                ]
            else:
                pseudo_dataset.teacher_data['weights'] = [
                    np.max(arr, axis=-1) for arr in teacher_predictions['proba']
                ]
                teacher_predictions['proba'] = None
        else:
            pseudo_dataset.teacher_data['weights'] = None
        pseudo_dataset.drop(col='labels', value=-1, type='teacher')
    
    def _update_pseudo_dataset_with_student(self, pseudo_dataset, student_predictions):
        """Update pseudo-dataset with student predictions."""
        self.logger.info("Update unlabeled data with Student's predictions")
        pseudo_dataset.student_data['id'] = student_predictions['id']
        pseudo_dataset.student_data['input_ids'] = student_predictions['input_ids']
        pseudo_dataset.student_data['align_index'] = student_predictions['align_index']
        pseudo_dataset.student_data['labels'] = student_predictions['preds']
        # If soft labels are required later, keep proba (downcast to float16 to save RAM);
        # otherwise, compute lightweight weights and drop proba to save memory
        if 'proba' in student_predictions and student_predictions['proba'] is not None:
            if self.config.model.soft_labels:
                pseudo_dataset.student_data['proba'] = [
                    arr.astype(np.float16) for arr in student_predictions['proba']
                ]
                pseudo_dataset.student_data['weights'] = [
                    np.max(arr, axis=-1) for arr in student_predictions['proba']
                ]
            else:
                pseudo_dataset.student_data['weights'] = [
                    np.max(arr, axis=-1) for arr in student_predictions['proba']
                ]
                student_predictions['proba'] = None
        else:
            pseudo_dataset.student_data['weights'] = None
        pseudo_dataset.drop(col='labels', value=-1, type='student')
    
    def _evaluate_student_iteration(self, datasets: Dict[str, Any], iteration: int) -> Dict[str, Any]:
        """Evaluate student after iteration."""
        # Dev evaluation
        self.logger.info("*** Evaluating student on dev data ***")
        dev_results = evaluate(
            self.student, datasets['dev'], self.evaluator,
            comment=f"student dev iter{iteration+1}"
        )
        self.logger.info(f"Student Dev performance on iter {iteration}: {dev_results['perf']}")
        
        # Test evaluation
        self.logger.info("*** Evaluating student on test data ***")
        test_results, test_predictions = evaluate(
            self.student, datasets['test'], self.evaluator,
            "test", comment=f"student test iter{iteration+1}"
        )
        self.logger.info(f"Student Test performance on iter {iteration}: {test_results['perf']}")
        
        self.performance_history['student_dev'].append(dev_results)
        self.performance_history['student_test'].append(test_results)
        
        return {
            'dev': dev_results,
            'test': test_results,
            'predictions': None
        }
    
    def _save_best_model_if_improved(self, dev_results: Dict[str, Any]):
        """Save best model if performance improved."""
        if not self.performance_history['student_dev']:
            return
        
        prev_max = max([x['perf'] for x in self.performance_history['student_dev'][:-1]])
        if dev_results['perf'] > prev_max:
            self.logger.info(f"Improved dev performance from {prev_max:.2f} to {dev_results['perf']:.2f}")
            self.student.save("student_best")
            if self.config.model.training_mode.value == 'weakly_supervised':
                self.teacher.save("teacher_best")
            # Save current test predictions immediately as best snapshot
            test_snapshot, test_preds = evaluate(
                self.student, self.data_handler.load_dataset(method='test'), self.evaluator,
                "test", comment="student test best_snapshot",
                remove_accents=self.config.training.remove_accents
            )
            write_predictions(
                self.config, self.logger, self.tokenizer,
                test_preds, file_name="student_best_predictions"
            )
    
    def finalize_results(self) -> Dict[str, Any]:
        """Finalize and save all results."""
        self.logger.info("Final Results")
        
        # Log teacher performances if applicable
        if self.config.model.training_mode.value == 'weakly_supervised':
            self._log_teacher_performances()
        
        # Log student performances
        self._log_student_performances()
        
        # Find best epoch
        best_dev_epoch, best_test_epoch = self._find_best_epochs()
        
        # Compile final results
        self.results = self._compile_final_results(best_dev_epoch, best_test_epoch)
        
        # Save predictions and models
        self._save_final_artifacts(best_dev_epoch)
        
        # Clean memory-heavy fields now that artifacts are persisted
        self._cleanup_memory()
        
        return self.results
    
    def _log_teacher_performances(self):
        """Log teacher performance history."""
        teacher_dev = [self._convert_numpy_types(x['perf']) for x in self.performance_history['teacher_dev']]
        teacher_test = [self._convert_numpy_types(x['perf']) for x in self.performance_history['teacher_test']]
        teacher_perf_str = [
            f"{i}:\t{teacher_dev[i]:.2f}\t{teacher_test[i]:.2f}" 
            for i in range(len(teacher_dev))
        ]
        self.logger.info("TEACHER PERFORMANCES:\n" + "\n".join(teacher_perf_str))
    
    def _log_student_performances(self):
        """Log student performance history."""
        student_dev = [self._convert_numpy_types(x['perf']) for x in self.performance_history['student_dev']]
        student_test = [self._convert_numpy_types(x['perf']) for x in self.performance_history['student_test']]
        student_perf_str = [
            f"{i}:\t{student_dev[i]:.2f}\t{student_test[i]:.2f}" 
            for i in range(len(student_dev))
        ]
        self.logger.info("STUDENT PERFORMANCES:\n" + "\n".join(student_perf_str))
    
    def _find_best_epochs(self) -> tuple:
        """Find best dev and test epochs."""
        dev_perfs = [self._convert_numpy_types(x['perf']) for x in self.performance_history['student_dev']]
        test_perfs = [self._convert_numpy_types(x['perf']) for x in self.performance_history['student_test']]
        
        best_dev_epoch = len(dev_perfs) - np.argmax(dev_perfs[::-1]) - 1
        best_test_epoch = len(test_perfs) - np.argmax(test_perfs[::-1]) - 1
        
        self.logger.info(f"BEST DEV {self.config.training.metric} = {dev_perfs[best_dev_epoch]:.3f} for epoch {best_dev_epoch}")
        self.logger.info(f"FINAL TEST {self.config.training.metric} = {test_perfs[best_dev_epoch]:.3f} for epoch {best_dev_epoch} (max={test_perfs[best_test_epoch]:.2f} for epoch {best_test_epoch})")
        
        return best_dev_epoch, best_test_epoch
    
    def _compile_final_results(self, best_dev_epoch: int, best_test_epoch: int) -> Dict[str, Any]:
        """Compile final results dictionary with clean data types and lightweight histories."""
        results = {}
        
        # Add iteration results (store lightweight summaries only)
        results['student_train_iter'] = self._extract_lightweight_history('student_train')
        results['student_dev_iter'] = self._extract_lightweight_history('student_dev')
        results['student_test_iter'] = self._extract_lightweight_history('student_test')
        
        # Add best results
        results['student_dev'] = self._convert_numpy_types(self.performance_history['student_dev'][best_dev_epoch])
        results['student_test'] = self._convert_numpy_types(self.performance_history['student_test'][best_dev_epoch])
        
        # Add teacher results if applicable
        if self.config.model.training_mode.value == 'weakly_supervised':
            results['teacher_train_iter'] = self._extract_lightweight_history('teacher_train')
            results['teacher_dev_iter'] = self._extract_lightweight_history('teacher_dev')
            results['teacher_test_iter'] = self._extract_lightweight_history('teacher_test')
            results['teacher_dev'] = self._convert_numpy_types(self.performance_history['teacher_dev'][best_dev_epoch])
            results['teacher_test'] = self._convert_numpy_types(self.performance_history['teacher_test'][best_dev_epoch])
        
        return results

    def _extract_lightweight_history(self, key: str) -> List[Dict[str, Any]]:
        """Keep only essential metrics to minimize memory in persisted results."""
        history = self.performance_history.get(key, [])
        lightweight: List[Dict[str, Any]] = []
        for entry in history:
            if isinstance(entry, dict):
                lightweight.append({
                    'precision': self._convert_numpy_types(entry.get('precision', 0.0)),
                    'recall': self._convert_numpy_types(entry.get('recall', 0.0)),
                    'normed_recall': self._convert_numpy_types(entry.get('normed_recall', 0.0)),
                    'f1_score': self._convert_numpy_types(entry.get('f1_score', 0.0)),
                    'accuracy': self._convert_numpy_types(entry.get('accuracy', 0.0)),
                    'perf': self._convert_numpy_types(entry.get('perf', 0.0)),
                })
            else:
                lightweight.append(self._convert_numpy_types(entry))
        return lightweight

    def _periodic_cleanup(self):
        """Trim in-memory artifacts during long runs to limit CPU RAM usage."""
        # Drop predictions entirely; we now stream them to disk
        self.student_predictions = []
        
        # For train histories, keep only the last 2 entries (dev/test are already lightweight)
        for key in ['student_train', 'teacher_train']:
            if key in self.performance_history and len(self.performance_history[key]) > 2:
                self.performance_history[key] = self.performance_history[key][-2:]

    def _cleanup_memory(self):
        """Free heavy objects after finalization to release RAM."""
        import gc
        # Already dropped predictions earlier; ensure cleared
        self.student_predictions = []
        # Train histories can be large; keep only the last summary
        for key in ['student_train', 'teacher_train']:
            if key in self.performance_history and len(self.performance_history[key]) > 1:
                self.performance_history[key] = [self.performance_history[key][-1]]
        gc.collect()
    
    def _convert_numpy_types(self, obj):
        """Convert numpy types to native Python types recursively."""
        if isinstance(obj, dict):
            return {key: self._convert_numpy_types(value) for key, value in obj.items()}
        elif isinstance(obj, list):
            return [self._convert_numpy_types(item) for item in obj]
        elif hasattr(obj, 'item'):  # numpy scalar
            return obj.item()
        elif hasattr(obj, 'tolist'):  # numpy array
            return obj.tolist()
        else:
            return obj
    
    def _save_final_artifacts(self, best_dev_epoch: int):
        """Save final predictions and models."""
        # Best predictions were already persisted when the best model was found.
        
        # Save final models
        self.student.save("student_last")
        if self.config.model.training_mode.value == 'weakly_supervised':
            self.teacher.save("teacher_last")
        
        # Save results
        save_and_report_results(self.config, self.results, self.logger)
